# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
High Level API for writting Tango clients

This is an experimental module. Not part of the official API.
"""

import weakref
import functools

try:
    from warnings import deprecated
except ImportError:
    from typing_extensions import deprecated

import tango
from tango import DeviceProxy as _DeviceProxy
from tango import CmdArgType
from tango.codec import loads
from tango.codec import dumps as _dumps

_FMT = "pickle"

dumps = functools.partial(_dumps, _FMT)


@deprecated(
    "Device alias was an experimental API - scheduled for removal in PyTango 11.0.0. "
    "Use tango.DeviceProxy instead."
)
class Device(_DeviceProxy):
    pass


def _command(device, cmd_info, *args, **kwargs):
    name = cmd_info.cmd_name
    if cmd_info.in_type == CmdArgType.DevEncoded:
        result = device.command_inout(name, dumps((args, kwargs)))
    else:
        result = device.command_inout(name, *args, **kwargs)
    if cmd_info.out_type == CmdArgType.DevEncoded:
        result = loads(*result)
    return result


class _DeviceHelper:
    __CMD_FILTER = {"init", "state", "status"}
    __ATTR_FILTER = {"state", "status"}

    __attr_cache = None
    __cmd_cache = None

    def __init__(self, dev_name, *args, **kwargs):
        self.dev_name = dev_name
        self.device = Device(dev_name, *args, **kwargs)
        self.slots = weakref.WeakKeyDictionary()

    def connect(self, signal, slot, event_type=tango.EventType.CHANGE_EVENT):
        i = self.device.subscribe_event(signal, event_type, slot)
        self.slots[slot] = i
        return i

    def disconnect(self, signal, slot):
        i = self.slots.pop(slot)
        self.device.unsubscribe_event(i)

    def get_attr_cache(self, refresh=False):
        cache = self.__attr_cache
        if not cache:
            refresh = True
        if refresh:
            cache = {}
            dev = self.device
            try:
                for attr_info in dev.attribute_list_query_ex():
                    attr_name = attr_info.name
                    if attr_name.lower() in self.__ATTR_FILTER:
                        continue
                    cache[attr_name] = attr_info
            except tango.DevFailed:
                pass
            self.__attr_cache = cache
        return cache

    def get_attr_info(self, name):
        cache = self.get_attr_cache()
        result = cache.get(name)
        if result:
            return result
        else:
            cache = self.get_attr_cache(refresh=True)
            return cache.get(name)

    def get_cmd_cache(self, refresh=False):
        cache = self.__cmd_cache
        if not cache:
            refresh = True
        if refresh:
            cache = {}
            dev = self.device
            try:
                for cmd_info in dev.command_list_query():
                    cmd_name = cmd_info.cmd_name
                    if cmd_name.lower() in self.__CMD_FILTER:
                        continue
                    cmd_func = functools.partial(_command, dev, cmd_info)
                    cmd_func.__name__ = cmd_name
                    cmd_func.__doc__ = cmd_info.in_type_desc
                    cmd_info.func = cmd_func
                    cache[cmd_name] = cmd_info
            except tango.DevFailed:
                pass
            self.__cmd_cache = cache
        return cache

    def get_cmd_info(self, name):
        cache = self.get_cmd_cache()
        result = cache.get(name)
        if result:
            return result
        else:
            cache = self.get_cmd_cache(refresh=True)
            return cache.get(name)

    def is_cmd(self, name):
        return name.lower() in self.get_cmd_cache()

    def members(self):
        result = self.get_attr_cache().keys()
        result.extend(self.get_cmd_cache().keys())
        return result

    def get(self, name):
        dev = self.device
        result = self.get_attr_info(name)
        if result:
            result = dev.read_attribute(name)
            value = result.value
            if result.type == tango.DevEncoded:
                result = loads(*value)
            else:
                result = value
            return result
        result = self.get_cmd_info(name)
        if result is None:
            raise KeyError(f"Unknown {name}")
        return result

    def set(self, name, value):
        result = self.get_attr_info(name)
        if result is None:
            raise KeyError(f"Unknown attribute {name}")
        if result.data_type == tango.DevEncoded:
            self.device.write_attribute(name, dumps(value))
        else:
            self.device.write_attribute(name, value)

    def get_info(self):
        try:
            return self.__info
        except AttributeError:
            pass
        try:
            info = self.device.info()
            self.__dict__["__info"] = info
            return info
        except tango.DevFailed:
            return None

    def __getitem__(self, name):
        if self.get_attr_info(name) is None:
            raise KeyError(f"Unknown attribute {name}")
        return self.device[name]

    def __setitem__(self, name, value):
        if self.get_attr_info(name) is None:
            raise KeyError(f"Unknown attribute {name}")
        self.device[name] = value

    def __str__(self):
        return self.dstr()

    def __repr__(self):
        return str(self)

    def dstr(self):
        info = self.get_info()
        klass = "Device"
        if info:
            klass = info.dev_class
        return f"{klass}({self.dev_name})"


@deprecated(
    "Object class was an experimental API - scheduled for removal in PyTango 11.0.0"
)
class Object:
    """Tango object"""

    def __init__(self, dev_name, *args, **kwargs):
        helper = _DeviceHelper(dev_name, *args, **kwargs)
        self.__dict__["_helper"] = helper

    def __getattr__(self, name):
        try:
            r = self._helper.get(name)
        except KeyError as ke:
            raise AttributeError(f"Unknown {name}") from ke
        if isinstance(r, tango.CommandInfo):
            self.__dict__[name] = r.func
            return r.func
        return r

    def __setattr__(self, name, value):
        try:
            return self._helper.set(name, value)
        except KeyError as ke:
            raise AttributeError(f"Unknown {name}") from ke

    def __getitem__(self, name):
        return self._helper[name]

    def __setitem__(self, name, value):
        self._helper[name] = value

    def __str__(self):
        return str(self._helper)

    def __repr__(self):
        return repr(self._helper)

    def __dir__(self):
        return self._helper.members()


@deprecated(
    "get_object_proxy function was an experimental API - scheduled for removal in PyTango 11.0.0"
)
def get_object_proxy(obj):
    """Experimental function. Not part of the official API"""
    return obj._helper.device


@deprecated(
    "get_object_db function was an experimental API - scheduled for removal in PyTango 11.0.0"
)
def get_object_db(obj):
    """Experimental function. Not part of the official API"""
    return get_object_proxy(obj).get_device_db()


@deprecated(
    "get_object_name function was an experimental API - scheduled for removal in PyTango 11.0.0"
)
def get_object_name(obj):
    """Experimental function. Not part of the official API"""
    return get_object_proxy(obj).get_name()


@deprecated(
    "get_object_info function was an experimental API - scheduled for removal in PyTango 11.0.0"
)
def get_object_info(obj):
    """Experimental function. Not part of the official API"""
    return get_object_proxy(obj).info()


@deprecated(
    "get_attributes_config function was an experimental API - scheduled for removal in PyTango 11.0.0"
)
def get_attributes_config(obj, refresh=False):
    """Experimental function. Not part of the official API"""
    return obj._helper.get_attr_cache(refresh=refresh)


@deprecated(
    "get_commands_config function was an experimental API - scheduled for removal in PyTango 11.0.0"
)
def get_commands_config(obj, refresh=False):
    """Experimental function. Not part of the official API"""
    return obj._helper.get_cmd_cache(refresh=refresh)


@deprecated(
    "connect function was an experimental API - scheduled for removal in PyTango 11.0.0"
)
def connect(obj, signal, slot, event_type=tango.EventType.CHANGE_EVENT):
    """Experimental function. Not part of the official API"""
    return obj._helper.connect(signal, slot, event_type=event_type)


@deprecated(
    "disconnect function was an experimental API - scheduled for removal in PyTango 11.0.0"
)
def disconnect(obj, signal, slot):
    """Experimental function. Not part of the official API"""
    return obj._helper.disconnect(signal, slot)



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""Define python methods for DeviceProxy object."""
from __future__ import annotations

import time
import textwrap
import threading
import functools
import enum
import collections.abc
import warnings
import collections
from typing import Any, Tuple, Dict

from tango.utils import PyTangoUserWarning

from tango._tango import StdStringVector, AttributeInfo, EventSubMode, DbDatum, DbData
from tango._tango import AttributeInfoEx, AttributeInfoList, AttributeInfoListEx
from tango._tango import DeviceProxy, __EventCallBack
from tango._tango import EventType, DevFailed, Except, ExtractAs, GreenMode
from tango._tango import constants
from tango._tango import (
    CmdArgType,
    DevState,
    DeviceAttribute,
    CommandInfoList,
    CommandInfo,
)

from tango.utils import is_pure_str, is_non_str_seq, is_integer
from tango.utils import seq_2_StdStringVector, StdStringVector_2_seq
from tango.utils import parameter_2_dbdata, get_property_from_db
from tango.utils import dir2
from tango.utils import _get_device_fqtrl_if_necessary
from tango.utils import _trace_client, _forcefully_traced_method
from tango.utils import _get_new_CallbackAutoDie
from tango.utils import _check_only_allowed_kwargs

from tango.green import green, green_callback
from tango.green import get_green_mode

__all__ = ("device_proxy_init", "get_device_proxy")

__docformat__ = "restructuredtext"

_UNSUBSCRIBE_LIFETIME = 60


@green(consume_green_mode=False)
@_trace_client
def get_device_proxy(
    *args, green_mode=None, executor=None, threadpool=None, asyncio_executor=None
):
    """get_device_proxy(self, dev_name, green_mode=None, wait=True, timeout=True) -> DeviceProxy
    get_device_proxy(self, dev_name, need_check_acc, green_mode=None, wait=True, timeout=None) -> DeviceProxy

    Returns a new :class:`~tango.DeviceProxy`.
    There is no difference between using this function and the direct
    :class:`~tango.DeviceProxy` constructor if you use the default kwargs.

    The added value of this function becomes evident when you choose a green_mode
    to be *Futures* or *Gevent* or *Asyncio*. The DeviceProxy constructor internally
    makes some network calls which makes it *slow*. By using one of the *green modes* as
    green_mode you are allowing other python code to be executed in a cooperative way.

    .. note::
        The timeout parameter has no relation with the tango device client side
        timeout (gettable by :meth:`~tango.DeviceProxy.get_timeout_millis` and
        settable through :meth:`~tango.DeviceProxy.set_timeout_millis`)

    :param dev_name: the device name or alias
    :type dev_name: str
    :param need_check_acc: in first version of the function it defaults to True.
                           Determines if at creation time of DeviceProxy it should check
                           for channel access (rarely used)
    :type need_check_acc: bool
    :param green_mode: determines the mode of execution of the device (including
                      the way it is created). Defaults to the current global
                      green_mode (check :func:`~tango.get_green_mode` and
                      :func:`~tango.set_green_mode`)
    :type green_mode: :obj:`~tango.GreenMode`
    :param wait: whether or not to wait for result. If green_mode
                 Ignored when green_mode is Synchronous (always waits).
    :type wait: bool
    :param timeout: The number of seconds to wait for the result.
                    If None, then there is no limit on the wait time.
                    Ignored when green_mode is Synchronous or wait is False.
    :type timeout: float
    :returns:
        if green_mode is Synchronous or wait is True:
            :class:`~tango.DeviceProxy`
        else if green_mode is Futures:
            :class:`concurrent.futures.Future`
        else if green_mode is Gevent:
            :class:`gevent.event.AsynchResult`
        else if green_mode is Asyncio:
            :class:`asyncio.Future`
    :throws:
        * a *DevFailed* if green_mode is Synchronous or wait is True
          and there is an error creating the device.
        * a *concurrent.futures.TimeoutError* if green_mode is Futures,
          wait is False, timeout is not None and the time to create the device
          has expired.
        * a *gevent.timeout.Timeout* if green_mode is Gevent, wait is False,
          timeout is not None and the time to create the device has expired.
        * a *asyncio.TimeoutError* if green_mode is Asyncio,
          wait is False, timeout is not None and the time to create the device
          has expired.

    New in PyTango 8.1.0
    """
    return DeviceProxy(
        *args,
        green_mode=green_mode,
        executor=executor,
        threadpool=threadpool,
        asyncio_executor=asyncio_executor,
    )


class __TangoInfo:
    """Helper class for copying DeviceInfo, or when DeviceProxy.info() fails."""

    def __init__(
        self,
        dev_class,
        dev_type,
        doc_url,
        server_host,
        server_id,
        server_version,
    ):
        self.dev_class = str(dev_class)
        self.dev_type = str(dev_type)
        self.doc_url = str(doc_url)
        self.server_host = str(server_host)
        self.server_id = str(server_id)
        self.server_version = int(server_version)

    @classmethod
    def from_defaults(cls):
        return cls(
            dev_class="Device",
            dev_type="Device",
            doc_url="Doc URL = https://www.tango-controls.org/developers/dsc",
            server_host="Unknown",
            server_id="Unknown",
            server_version=1,
        )

    @classmethod
    def from_copy(cls, info):
        return cls(
            dev_class=info.dev_class,
            dev_type=info.dev_type,
            doc_url=info.doc_url,
            server_host=info.server_host,
            server_id=info.server_id,
            server_version=info.server_version,
        )


# -------------------------------------------------------------------------------
# Pythonic API: transform tango commands into methods and tango attributes into
# class members
# -------------------------------------------------------------------------------


def __check_read_attribute(dev_attr):
    if dev_attr.has_failed:
        raise DevFailed(*dev_attr.get_err_stack())
    return dev_attr


def __init_device_proxy_internals(proxy):
    if proxy.__dict__.get("_initialized", False):
        return
    executors = {key: None for key in GreenMode.values.values()}
    proxy.__dict__["_green_mode"] = None
    proxy.__dict__["_dynamic_interface_frozen"] = True
    proxy.__dict__["_initialized"] = True
    proxy.__dict__["_executors"] = executors
    proxy.__dict__["_pending_unsubscribe"] = {}


def __DeviceProxy__get_cmd_cache(self):
    try:
        ret = self.__dict__["__cmd_cache"]
    except KeyError:
        self.__dict__["__cmd_cache"] = ret = {}
    return ret


def __DeviceProxy__get_attr_cache(self):
    try:
        ret = self.__dict__["__attr_cache"]
    except KeyError:
        self.__dict__["__attr_cache"] = ret = {}
    return ret


def __DeviceProxy____init__(
    self, *args, green_mode=None, executor=None, threadpool=None, asyncio_executor=None
):
    __init_device_proxy_internals(self)
    self.__dict__["_initialized"] = False
    self.__dict__["_green_mode"] = green_mode
    self.__dict__["_executors"][GreenMode.Futures] = executor
    self.__dict__["_executors"][GreenMode.Gevent] = threadpool
    self.__dict__["_executors"][GreenMode.Asyncio] = asyncio_executor

    # If TestContext active, short TRL is replaced with fully-qualified
    # TRL, using test server's connection details.  Otherwise, left as-is.
    device_name = args[0]
    new_device_name = _get_device_fqtrl_if_necessary(device_name)
    new_args = [new_device_name] + list(args[1:])
    try:
        DeviceProxy.__init_orig__(self, *new_args)
    except DevFailed as orig_err:
        if new_device_name != device_name:
            # If device was not found, it could be an attempt to access a real device
            # with short name while running TestContext.  I.e., we need to use the
            # short name so that the real TANGO_HOST will be tried.
            try:
                DeviceProxy.__init_orig__(self, *args)
            except DevFailed as retry_exc:
                Except.re_throw_exception(
                    retry_exc,
                    "PyAPI_DeviceProxyInitFailed",
                    f"Failed to create DeviceProxy "
                    f"(tried {new_device_name!r} => {orig_err.args[0].reason}, and "
                    f"{device_name!r} => {retry_exc.args[0].reason})",
                    "__DeviceProxy__init__",
                )
        else:
            raise

    self.__dict__["_initialized"] = True


def __DeviceProxy__get_green_mode(self):
    """Returns the green mode in use by this DeviceProxy.

    :returns: the green mode in use by this DeviceProxy.
    :rtype: GreenMode

    .. seealso::
        :func:`tango.get_green_mode`
        :func:`tango.set_green_mode`

    New in PyTango 8.1.0
    """
    gm = self._green_mode
    if gm is None:
        gm = get_green_mode()
    return gm


def __DeviceProxy__set_green_mode(self, green_mode=None):
    """Sets the green mode to be used by this DeviceProxy
    Setting it to None means use the global PyTango green mode
    (see :func:`tango.get_green_mode`).

    :param green_mode: the new green mode
    :type green_mode: GreenMode

    New in PyTango 8.1.0
    """
    self._green_mode = green_mode


def __DeviceProxy__refresh_cmd_cache(self):
    cmd_list = self.command_list_query()
    cmd_cache = {}
    for cmd in cmd_list:
        n = cmd.cmd_name.lower()
        doc = f"{cmd.cmd_name}({cmd.in_type}) -> {cmd.out_type}\n\n"
        doc += f" -  in ({cmd.in_type}): {cmd.in_type_desc}\n"
        doc += f" - out ({cmd.out_type}): {cmd.out_type_desc}\n"
        cmd_cache[n] = cmd, doc
    self.__dict__["__cmd_cache"] = cmd_cache


def __DeviceProxy__refresh_attr_cache(self):
    attr_list = self.attribute_list_query_ex()
    attr_cache = {}
    for attr in attr_list:
        name = attr.name.lower()
        enum_class = None
        if attr.data_type == CmdArgType.DevEnum and attr.enum_labels:
            enum_class = enum.IntEnum(attr.name, " ".join(attr.enum_labels), start=0)
        elif attr.data_type == CmdArgType.DevState:
            enum_class = DevState
        attr_cache[name] = (
            attr.name,
            enum_class,
        )
    self.__dict__["__attr_cache"] = attr_cache


def __DeviceProxy__freeze_dynamic_interface(self):
    """Prevent unknown attributes to be set on this DeviceProxy instance.

    An exception will be raised if the Python attribute set on this DeviceProxy
    instance does not already exist.  This prevents accidentally writing to
    a non-existent Tango attribute when using the high-level API.

    This is the default behaviour since PyTango 9.3.4.

    See also :meth:`tango.DeviceProxy.unfreeze_dynamic_interface`.

    .. versionadded:: 9.4.0
    """
    self._dynamic_interface_frozen = True


def __DeviceProxy__unfreeze_dynamic_interface(self):
    """Allow new attributes to be set on this DeviceProxy instance.

    An exception will not be raised if the Python attribute set on this DeviceProxy
    instance does not exist.  Instead, the new Python attribute will be added to
    the DeviceProxy instance's dictionary of attributes.  This may be useful, but
    a user will not get an error if they accidentally write to a non-existent Tango
    attribute when using the high-level API.

    See also :meth:`tango.DeviceProxy.freeze_dynamic_interface`.

    .. versionadded:: 9.4.0
    """
    warnings.warn(
        f"Dynamic interface unfrozen on DeviceProxy instance {self} id=0x{id(self):x} - "
        f"arbitrary Python attributes can be set without raising an exception.",
        category=PyTangoUserWarning,
    )
    self._dynamic_interface_frozen = False


def __DeviceProxy__is_dynamic_interface_frozen(self):
    """Indicates if the dynamic interface for this DeviceProxy instance is frozen.

    See also :meth:`tango.DeviceProxy.freeze_dynamic_interface` and
    :meth:`tango.DeviceProxy.unfreeze_dynamic_interface`.

        :returns: True if the dynamic interface this DeviceProxy is frozen.
        :rtype: bool

    .. versionadded:: 9.4.0
    """
    return self._dynamic_interface_frozen


def __get_command_func(dp, cmd_info, name):
    _, doc = cmd_info

    def f(*args, **kwds):
        return dp.command_inout(name, *args, **kwds)

    f.__doc__ = doc
    return f


def __update_enum_values(attr_info, attr_value):
    _, enum_class = attr_info
    if enum_class and attr_value is not None:
        if is_non_str_seq(attr_value):
            ret = []
            for value in attr_value:
                if is_non_str_seq(value):
                    ret.append(tuple([enum_class(v) for v in value]))
                else:
                    ret.append(enum_class(value))
            return tuple(ret)

        return enum_class(attr_value)
    else:
        return attr_value


async def __async_get_attribute_value(self, attr_info, name):
    reading = await self.read_attribute(name)
    return __update_enum_values(attr_info, reading.value)


def __sync_get_attribute_value(self, attr_info, name):
    reading = self.read_attribute(name)
    return __update_enum_values(attr_info, reading.value)


def __get_attribute_value(self, attr_info, name):
    if self.get_green_mode() == GreenMode.Asyncio:
        return __async_get_attribute_value(self, attr_info, name)
    else:
        return __sync_get_attribute_value(self, attr_info, name)


def __convert_str_to_enum(value, enum_class, attr_name):
    try:
        return enum_class[value]
    except KeyError:
        raise AttributeError(
            f"Invalid enum value {value} for attribute {attr_name}. "
            f"Valid values: {[m for m in enum_class.__members__.keys()]}"
        )


def __set_attribute_value(self, name, value):
    attr_info = self.__get_attr_cache().get(name.lower())
    if attr_info:
        # allow writing DevEnum attributes using string values
        _, enum_class = attr_info
        if enum_class:
            if is_non_str_seq(value):
                org_value = value
                value = []
                for val in org_value:
                    if is_non_str_seq(val):
                        value.append(
                            [
                                (
                                    __convert_str_to_enum(v, enum_class, name)
                                    if is_pure_str(v)
                                    else v
                                )
                                for v in val
                            ]
                        )
                    else:
                        value.append(
                            __convert_str_to_enum(val, enum_class, name)
                            if is_pure_str(val)
                            else val
                        )
            elif is_pure_str(value):
                value = __convert_str_to_enum(value, enum_class, name)
    return self.write_attribute(name, value)


def __DeviceProxy__getattr(self, name):
    cause = None
    try:
        # trait_names is a feature of IPython. Hopefully they will solve
        # ticket http://ipython.scipy.org/ipython/ipython/ticket/229 someday
        # and the ugly trait_names could be removed.
        if name.startswith("_") or name == "trait_names":
            raise AttributeError(name) from cause

        name_l = name.lower()

        cmd_info = self.__get_cmd_cache().get(name_l)
        if cmd_info:
            return __get_command_func(self, cmd_info, name)

        attr_info = self.__get_attr_cache().get(name_l)
        if attr_info:
            try:
                return __get_attribute_value(self, attr_info, name)
            except DevFailed as err:
                # it could be, that attribute was deleted and we have to re-create attribute cache
                if err.args[0].reason != "API_AttrNotFound":
                    raise

        try:
            self.__refresh_cmd_cache()
        except Exception as e:
            if cause is None:
                cause = e

        cmd_info = self.__get_cmd_cache().get(name_l)
        if cmd_info:
            return __get_command_func(self, cmd_info, name)

        try:
            self.__refresh_attr_cache()
        except Exception as e:
            if cause is None:
                cause = e

        attr_info = self.__get_attr_cache().get(name_l)
        if attr_info:
            return __get_attribute_value(self, attr_info, name)

        raise AttributeError(name) from cause
    finally:
        del cause


def __DeviceProxy__setattr(self, name, value):
    cause = None
    try:
        name_l = name.lower()

        if name_l in self.__get_cmd_cache():
            raise TypeError("Cannot set the value of a command") from cause

        if name_l in self.__get_attr_cache():
            return __set_attribute_value(self, name, value)

        try:
            self.__refresh_cmd_cache()
        except Exception as e:
            if cause is None:
                cause = e

        if name_l in self.__get_cmd_cache():
            raise TypeError("Cannot set the value of a command") from cause

        try:
            self.__refresh_attr_cache()
        except Exception as e:
            if cause is None:
                cause = e

        if name_l in self.__get_attr_cache():
            return __set_attribute_value(self, name, value)

        try:
            if name in self.__dict__ or not self.is_dynamic_interface_frozen():
                return super(DeviceProxy, self).__setattr__(name, value)
            else:
                raise AttributeError(
                    f"Tried to set non-existent attr {repr(name)} to {repr(value)}.\n"
                    f"The DeviceProxy object interface is frozen and cannot be modified - "
                    f"see tango.DeviceProxy.freeze_dynamic_interface for details."
                )
        except Exception as e:
            raise e from cause
    finally:
        del cause


def __DeviceProxy__dir(self):
    """Return the attribute list including tango objects."""
    extra_entries = set()
    # Add commands
    try:
        extra_entries.update(self.get_command_list())
    except Exception:
        pass
    # Add attributes
    try:
        extra_entries.update(self.get_attribute_list())
    except Exception:
        pass
    # Merge with default dir implementation
    extra_entries.update([x.lower() for x in extra_entries])
    entries = extra_entries.union(dir2(self))
    return sorted(entries)


def __DeviceProxy__getitem(self, key):
    return self.read_attribute(key)


def __DeviceProxy__setitem(self, key, value):
    return self.write_attribute(key, value)


def __DeviceProxy__contains(self, key):
    return key.lower() in map(str.lower, self.get_attribute_list())


def __DeviceProxy__read_attribute(
    self, value, extract_as=ExtractAs.Numpy
) -> DeviceAttribute:
    """
    read_attribute(self, value, extract_as=ExtractAs.Numpy, __GREEN_KWARGS__) -> DeviceAttribute

        Read a single attribute.

    :param value: The name of the attribute to read.
    :type value: str

    :param extract_as: Defaults to numpy.
    :type extract_as: :obj:`tango.ExtractAs`

    __GREEN_KWARGS_DESCRIPTION__

    :return: DeviceAttribute object with read attribute value.
    :rtype: :obj:`tango.DeviceAttribute`

    :throws: :obj:`tango.ConnectionFailed`: Raised in case of a connection failure.
    :throws: :obj:`tango.CommunicationFailed`: Raised in case of a communication failure.
    :throws: :obj:`tango.DevFailed`: Raised in case of a device failure.
    __GREEN_RAISES__

    .. versionchanged:: 7.1.4
        For DevEncoded attributes, before it was returning a DeviceAttribute.value
        as a tuple **(format<str>, data<str>)** no matter what was the *extract_as*
        value was. Since 7.1.4, it returns a **(format<str>, data<buffer>)**
        unless *extract_as* is String, in which case it returns
        **(format<str>, data<str>)**.

    .. versionchanged:: 8.0.0
        For DevEncoded attributes, now returns a DeviceAttribute.value
        as a tuple **(format<str>, data<bytes>)** unless *extract_as* is String,
        in which case it returns **(format<str>, data<str>)**. Careful, if
        using python >= 3 data<str> is decoded using default python
        *utf-8* encoding. This means that PyTango assumes tango DS was written
        encapsulating string into *utf-8* which is the default python encoding.

    .. versionadded:: 8.1.0
        *green_mode* parameter.
        *wait* parameter.
        *timeout* parameter.

    .. versionchanged:: 9.4.0
        For spectrum and image attributes with an empty sequence, no longer
        returns DeviceAttribute.value and DeviceAttribute.w_value as
        :obj:`None`.  Instead, DevString and DevEnum types get an empty :obj:`tuple`,
        while other types get an empty :obj:`numpy.ndarray`.  Using *extract_as* can
        change the sequence type, but it still won't be :obj:`None`.
    """
    return __check_read_attribute(self._read_attribute(value, extract_as))


def __read_attributes_asynch__(self, attr_names, cb, extract_as):
    if cb is None:
        return self.__read_attributes_asynch(attr_names)

    cb2 = _get_new_CallbackAutoDie()
    if isinstance(cb, collections.abc.Callable):
        cb2.attr_read = _forcefully_traced_method(cb)
    else:
        cb2.attr_read = _forcefully_traced_method(cb.attr_read)
    return self.__read_attributes_asynch(attr_names, cb2, extract_as)


def __DeviceProxy__read_attributes_asynch(
    self, attr_names, cb=None, extract_as=ExtractAs.Numpy
) -> int | None:
    """
    read_attributes_asynch(self, attr_names, __GREEN_KWARGS__) -> int
    read_attributes_asynch(self, attr_names, cb, extract_as=Numpy, __GREEN_KWARGS__) -> None

        Read asynchronously an attribute list.

        New in PyTango 7.0.0

    .. important::
        by default, TANGO is initialized with the **polling** model. If you want
        to use the **push** model (the one with the callback parameter), you
        need to change the global TANGO model to PUSH_CALLBACK.
        You can do this with the :meth:`tango.ApiUtil.set_asynch_cb_sub_model`

    :param attr_names: A list of attributes to read. See read_attributes.
    :type attr_names: Sequence[str]

    :param cb: push model: as soon as attributes read, core calls cb with read results.
        This callback object should be an instance of a user class with an attr_read() method.
        It can also be any callable object.
    :type cb: Optional[Callable]

    :param extract_as: Defaults to numpy.
    :type extract_as: :obj:`tango.ExtractAs`

    __GREEN_KWARGS_DESCRIPTION__

    :returns: an asynchronous call identifier which is needed to get attributes value if poll model, None if push model
    :rtype: Union[int, None]

    :throws: :obj:`tango.ConnectionFailed`
    __GREEN_RAISES__

    .. important::
        Multiple asynchronous calls are not guaranteed to be executed by the device
        server in the same order they are invoked by the client.  E.g., a call
        to ``write_attributes_asynch([("a", 1)])`` followed immediately with a call to
        ``read_attributes_asynch(["a"])`` could result in the device reading the
        attribute ``a`` before writing to it.
    """

    return __read_attributes_asynch__(self, attr_names, cb, extract_as)


def __DeviceProxy__read_attribute_asynch(
    self, attr_name, cb=None, extract_as=ExtractAs.Numpy
) -> int | None:
    """
    read_attribute_asynch(self, attr_name, __GREEN_KWARGS__) -> int
    read_attribute_asynch(self, attr_name, cb, extract_as=Numpy, __GREEN_KWARGS__) -> None

        Read asynchronously the specified attributes.

        New in PyTango 7.0.0

    .. important::
        by default, TANGO is initialized with the **polling** model. If you want
        to use the **push** model (the one with the callback parameter), you
        need to change the global TANGO model to PUSH_CALLBACK.
        You can do this with the :meth:`tango.ApiUtil.set_asynch_cb_sub_model`

    :param attr_name: an attribute to read
    :type attr_name: str

    :param cb: push model: as soon as attributes read, core calls cb with read results.
        This callback object should be an instance of a user class with an attr_read() method.
        It can also be any callable object.
    :type cb: Optional[Callable]

    :param extract_as: Defaults to numpy.
    :type extract_as: ExtractAs

    __GREEN_KWARGS_DESCRIPTION__

    :returns: an asynchronous call identifier which is needed to get attribute value if poll model, None if push model
    :rtype: Union[int, None]

    :throws: :obj:`tango.ConnectionFailed`
    __GREEN_RAISES__

    .. important::
        Multiple asynchronous calls are not guaranteed to be executed by the device
        server in the same order they are invoked by the client.  E.g., a call
        to the method ``write_attribute_asynch("a", 1)`` followed immediately with
        a call to ``read_attribute_asynch("a")`` could result in the device reading the
        attribute ``a`` before writing to it.
    """
    return __read_attributes_asynch__(self, [attr_name], cb, extract_as)


def __read_attributes_reply__(self, *args, **kwargs):
    if "poll_timeout" in kwargs:
        kwargs["timeout"] = kwargs.pop("poll_timeout")

    return self.__read_attributes_reply(*args, **kwargs)


def __DeviceProxy__read_attributes_reply(
    self, *args, **kwargs
) -> list[DeviceAttribute]:
    """
    read_attributes_reply(self, id, extract_as=ExtractAs.Numpy, __GREEN_KWARGS__) -> [DeviceAttribute]
    read_attributes_reply(self, id, poll_timeout, extract_as=ExtractAs.Numpy, __GREEN_KWARGS__) -> [DeviceAttribute]

    Get the answer of an asynchronous read_attributes call, if it has arrived (polling model).

    If the reply is ready, but an attribute raised an exception while reading, it will
    still be included in the returned list.  However, the has_error field for that item
    will be set to True.

    .. versionchanged:: 7.0.0 New in PyTango
    .. versionchanged:: 10.0.0 To eliminate confusion between different timeout parameters, the core (cppTango) timeout (previously the optional second positional argument) has been renamed to "poll_timeout". Conversely, the pyTango executor timeout remains as the keyword argument "timeout". These parameters have distinct meanings and units:

        - The cppTango "poll_timeout" is measured in milliseconds and blocks the call until a reply is received. If the reply is not received within the specified poll_timeout duration, an exception is thrown. Setting poll_timeout to 0 causes the call to wait indefinitely until a reply is received.
        - The pyTango "timeout" is measured in seconds and is applicable only in asynchronous GreenModes (Asyncio, Futures, Gevent), and only when "wait" is set to True. The specific behavior when a reply is not received within the specified timeout period varies depending on the GreenMode.


    :param id: the asynchronous call identifier
    :type id: int

    :param poll_timeout: cppTango core timeout in ms.
        If the reply has not yet arrived, the call will wait for the time specified (in ms).
        If after timeout, the reply is still not there, an exception is thrown.
        If timeout set to 0, the call waits until the reply arrives.
        If the argument is not provided, then there is no timeout check, and an
        exception is raised immediately if the reply is not ready.
    :type poll_timeout: Optional[int]

    :param extract_as: Defaults to numpy.
    :type extract_as: ExtractAs

    __GREEN_KWARGS_DESCRIPTION__

    :returns: If the reply is arrived and if it is a valid reply,
        it is returned to the caller in a list of DeviceAttribute.
        If the reply is an exception, it is re-thrown by this call.
        If the reply is not yet arrived, the call will wait (blocking the process)
        for the time specified in timeout. If after timeout milliseconds, the reply is still not there, an
        exception is thrown. If timeout is set to 0, the call waits
        until the reply arrived.
    :rtype: Sequence[DeviceAttribute]

    :throws: Union[AsynCall, AsynReplyNotArrived, ConnectionFailed, CommunicationFailed, DevFailed]
    __GREEN_RAISES__
    """
    _check_only_allowed_kwargs(kwargs, {"id", "poll_timeout", "extract_as"})

    return __read_attributes_reply__(self, *args, **kwargs)


def __DeviceProxy__read_attribute_reply(self, *args, **kwargs) -> DeviceAttribute:
    """
    read_attribute_reply(self, id, extract_as=ExtractAs.Numpy, __GREEN_KWARGS__) -> DeviceAttribute
    read_attribute_reply(self, id, poll_timeout, extract_as=ExtractAs.Numpy, __GREEN_KWARGS__) -> DeviceAttribute

    Get the answer of an asynchronous read_attribute call, if it has arrived (polling model).

    If the reply is ready, but the attribute raised an exception while reading, an
    exception will be raised by this function (DevFailed, with reason API_AttrValueNotSet).

    .. versionchanged:: 7.0.0 New in PyTango
    .. versionchanged:: 10.0.0 To eliminate confusion between different timeout parameters, the core (cppTango) timeout (previously the optional second positional argument) has been renamed to "poll_timeout". Conversely, the pyTango executor timeout remains as the keyword argument "timeout". These parameters have distinct meanings and units:

        - The cppTango "poll_timeout" is measured in milliseconds and blocks the call until a reply is received. If the reply is not received within the specified poll_timeout duration, an exception is thrown. Setting poll_timeout to 0 causes the call to wait indefinitely until a reply is received.
        - The pyTango "timeout" is measured in seconds and is applicable only in asynchronous GreenModes (Asyncio, Futures, Gevent), and only when "wait" is set to True. The specific behavior when a reply is not received within the specified timeout period varies depending on the GreenMode.


    :param id: the asynchronous call identifier
    :type id: int

    :param poll_timeout: cppTango core timeout in ms.
        If the reply has not yet arrived, the call will wait for the time specified (in ms).
        If after timeout, the reply is still not there, an exception is thrown.
        If timeout set to 0, the call waits until the reply arrives.
        If the argument is not provided, then there is no timeout check, and an
        exception is raised immediately if the reply is not ready.
    :type poll_timeout: Optional[int]

    :param extract_as: Defaults to numpy.
    :type extract_as: ExtractAs

    __GREEN_KWARGS_DESCRIPTION__

    :returns: If the reply is arrived and if it is a valid reply,
        it is returned to the caller in a list of DeviceAttribute.
        If the reply is an exception, it is re-thrown by this call.
        If the reply is not yet arrived, the call will wait (blocking the process)
        for the time specified in timeout. If after timeout milliseconds, the reply is still not there, an
        exception is thrown. If timeout is set to 0, the call waits
        until the reply arrived.
    :rtype: :obj:`tango.DeviceAttribute`

    :throws: Union[AsynCall, AsynReplyNotArrived, ConnectionFailed, CommunicationFailed, DevFailed]

    """
    _check_only_allowed_kwargs(kwargs, {"id", "poll_timeout", "extract_as"})

    attr = __read_attributes_reply__(self, *args, **kwargs)[0]
    return __check_read_attribute(attr)


def __write_attributes_asynch__(self, attr_values, cb=None):
    if cb is None:
        return self.__write_attributes_asynch(attr_values)

    cb2 = _get_new_CallbackAutoDie()
    if isinstance(cb, collections.abc.Callable):
        cb2.attr_written = _forcefully_traced_method(cb)
    else:
        cb2.attr_written = _forcefully_traced_method(cb.attr_written)
    return self.__write_attributes_asynch(attr_values, cb2)


def __DeviceProxy__write_attributes_asynch(
    self, attr_values: list[tuple[str | AttributeInfoEx, Any]], cb=None
) -> int | None:
    """
    Write asynchronously the specified attributes.

    .. important::
        by default, TANGO is initialized with the **polling** model. If you want
        to use the **push** model (the one with the callback parameter), you
        need to change the global TANGO model to PUSH_CALLBACK.
        You can do this with the :meth:`tango.ApiUtil.set_asynch_cb_sub_model`

    :param attr_values: pairs of (attr_name, value) to write (see Note below)
    :type attr_values: Sequence[Sequence[str | :obj:`~tango.AttributeInfoEx`, Any]]

    :param cb: push model: as soon as attributes written, core calls cb with write results.
        This callback object should be an instance of a user class with an attr_written() method.
        It can also be any callable object.
    :type cb: Optional[Callable]

    __GREEN_KWARGS_DESCRIPTION__

    :returns: an asynchronous call identifier which is needed to get the server reply if poll model, None if push model
    :rtype: Union[int, None]

    :throws: :obj:`tango.ConnectionFailed`
    __GREEN_RAISES__

    .. important::
        Multiple asynchronous calls are not guaranteed to be executed by the device
        server in the same order they are invoked by the client.  E.g., a call
        to ``write_attributes_asynch([("a", 1)])`` followed immediately with a call to
        ``read_attributes_asynch(["a"])`` could result in the device reading the
        attribute ``a`` before writing to it.

    .. note::
        For each pair of values there are two possibilities for the
        attr parameter: if you give attribute name, then PyTango must
        fetch attribute info for this attribute from server by additional synchronous(!)
        IO, since we must know to which c++ data type cast each python value.
        If you would like to avoid this IO you must give AttributeInfoEx instead of
        attribute name for each(!) pair of values.

    .. versionchanged:: 10.1.0 Added support for AttributeInfoEx for attr_values parameter.

    """

    return __write_attributes_asynch__(self, attr_values, cb)


def __DeviceProxy__write_attribute_asynch(
    self, attr: str | AttributeInfoEx, value: Any, cb=None
) -> int | None:
    """
    Write asynchronously the specified attribute.

    .. important::
        by default, TANGO is initialized with the **polling** model. If you want
        to use the **push** model (the one with the callback parameter), you
        need to change the global TANGO model to PUSH_CALLBACK.
        You can do this with the :meth:`tango.ApiUtil.set_asynch_cb_sub_model`

    :param attr: an attribute name to write or AttributeInfoEx object (see Note below)
    :type attr: str | :obj:`~tango.AttributeInfoEx`

    :param value: value to write
    :type value: Any

    :param cb: push model: as soon as attribute written, core calls cb with write results.
        This callback object should be an instance of a user class with an attr_written() method.
        It can also be any callable object.
    :type cb: Optional[Callable]

    __GREEN_KWARGS_DESCRIPTION__

    :returns: an asynchronous call identifier which is needed to get the server reply if poll model, None if push model
    :rtype: Union[int, None]

    :throws: ConnectionFailed
    __GREEN_RAISES__

    .. important::
        Multiple asynchronous calls are not guaranteed to be executed by the device
        server in the same order they are invoked by the client.  E.g., a call
        to the method ``write_attribute_asynch("a", 1)`` followed immediately with
        a call to ``read_attribute_asynch("a")`` could result in the device reading the
        attribute ``a`` before writing to it.

    .. note::
        There are two possibilities for the attr parameter: if you give attribute name,
        then PyTango will do additional synchronous(!) IO to fetch AttributeInfoEx object from device server,
        since we must know to which c++ data type cast python value. If you would like to avoid this IO you
        can give AttributeInfoEx instead of attribute name.

    .. versionchanged:: 10.1.0

        - attr_name parameter was renamed to attr
        - added support for AttributeInfoEx for attr parameter

    """
    return __write_attributes_asynch__(self, [(attr, value)], cb)


def __write_attributes_reply__(self, *args, **kwargs):
    if "poll_timeout" in kwargs:
        kwargs["timeout"] = kwargs.pop("poll_timeout")

    return self.__write_attributes_reply(*args, **kwargs)


def __DeviceProxy__write_attributes_reply(self, *args, **kwargs) -> None:
    """

    write_attributes_reply(self, id, __GREEN_KWARGS__) -> None
    write_attributes_reply(self, id, poll_timeout, __GREEN_KWARGS__) -> None

        Check if the answer of an asynchronous write_attributes is arrived
        (polling model). If the reply is arrived and if it is a valid reply,
        the call returned. If the reply is an exception, it is re-thrown by
        this call. An exception is also thrown in case of the reply is not
        yet arrived.

    .. versionchanged:: 7.0.0 New in PyTango
    .. versionchanged:: 10.0.0 To eliminate confusion between different timeout parameters, the core (cppTango) timeout (previously the optional second positional argument) has been renamed to "poll_timeout". Conversely, the pyTango executor timeout remains as the keyword argument "timeout". These parameters have distinct meanings and units:

        - The cppTango "poll_timeout" is measured in milliseconds and blocks the call until a reply is received. If the reply is not received within the specified poll_timeout duration, an exception is thrown. Setting poll_timeout to 0 causes the call to wait indefinitely until a reply is received.
        - The pyTango "timeout" is measured in seconds and is applicable only in asynchronous GreenModes (Asyncio, Futures, Gevent), and only when "wait" is set to True. The specific behavior when a reply is not received within the specified timeout period varies depending on the GreenMode.

    :param id: the asynchronous call identifier
    :type id: int

    :param poll_timeout: cppTango core timeout in ms.
        If the reply has not yet arrived, the call will wait for the time specified (in ms).
        If after timeout, the reply is still not there, an exception is thrown.
        If timeout set to 0, the call waits until the reply arrives.
        If the argument is not provided, then there is no timeout check, and an
        exception is raised immediately if the reply is not ready.
    :type poll_timeout: Optional[int]

    :param extract_as: Defaults to numpy.
    :type extract_as: ExtractAs

    __GREEN_KWARGS_DESCRIPTION__

    :returns: None
    :rtype: None

    :throws: Union[AsynCall, AsynReplyNotArrived, ConnectionFailed, CommunicationFailed, DevFailed]
    __GREEN_RAISES__

    """
    _check_only_allowed_kwargs(kwargs, {"id", "poll_timeout"})

    return __write_attributes_reply__(self, *args, **kwargs)


def __DeviceProxy__write_attribute_reply(self, *args, **kwargs) -> None:
    """
    write_attribute_reply(self, id, __GREEN_KWARGS__) -> None
    write_attribute_reply(self, id, poll_timeout, __GREEN_KWARGS__) -> None

        Check if the answer of an asynchronous write_attributes is arrived
        (polling model). If the reply is arrived and if it is a valid reply,
        the call returned. If the reply is an exception, it is re-thrown by
        this call. An exception is also thrown in case of the reply is not
        yet arrived.

    .. versionchanged:: 7.0.0 New in PyTango
    .. versionchanged:: 10.0.0 To eliminate confusion between different timeout parameters, the core (cppTango) timeout (previously the optional second positional argument) has been renamed to "poll_timeout". Conversely, the pyTango executor timeout remains as the keyword argument "timeout". These parameters have distinct meanings and units:

        - The cppTango "poll_timeout" is measured in milliseconds and blocks the call until a reply is received. If the reply is not received within the specified poll_timeout duration, an exception is thrown. Setting poll_timeout to 0 causes the call to wait indefinitely until a reply is received.
        - The pyTango "timeout" is measured in seconds and is applicable only in asynchronous GreenModes (Asyncio, Futures, Gevent), and only when "wait" is set to True. The specific behavior when a reply is not received within the specified timeout period varies depending on the GreenMode.

    :param id: the asynchronous call identifier
    :type id: int

    :param poll_timeout: cppTango core timeout in ms.
        If the reply has not yet arrived, the call will wait for the time specified (in ms).
        If after timeout, the reply is still not there, an exception is thrown.
        If timeout set to 0, the call waits until the reply arrives.
        If the argument is not provided, then there is no timeout check, and an
        exception is raised immediately if the reply is not ready.
    :type poll_timeout: Optional[int]

    :param extract_as: Defaults to numpy.
    :type extract_as: ExtractAs

    __GREEN_KWARGS_DESCRIPTION__

    :returns: None
    :rtype: None

    :throws: Union[AsynCall, AsynReplyNotArrived, ConnectionFailed, CommunicationFailed, DevFailed]
    """
    _check_only_allowed_kwargs(kwargs, {"id", "poll_timeout"})

    return __write_attributes_reply__(self, *args, **kwargs)


def __DeviceProxy__write_read_attribute(
    self,
    attr: str | AttributeInfoEx,
    value: Any,
    extract_as: ExtractAs = ExtractAs.Numpy,
) -> DeviceAttribute:
    """

     Write then read a single attribute in a single network call.
     By default (serialisation by device), the execution of this call in
     the server can't be interrupted by other clients.

     :param attr: The name or AttributeInfoEx structure of the attribute to write.
     :type attr: str | :obj:`~tango.AttributeInfoEx`

     :param value: The value to write to the attribute.
     :type value: Any

     __GREEN_KWARGS_DESCRIPTION__

     :returns: A DeviceAttribute object with readout value.
     :rtype: :obj:`tango.DeviceAttribute`

     :throws: :obj:`tango.ConnectionFailed`: Raised in case of a connection failure.
     :throws: :obj:`tango.CommunicationFailed`: Raised in case of a communication failure.
     :throws: :obj:`tango.DeviceUnlocked`: Raised in case of an unlocked device.
     :throws: :obj:`tango.WrongData`: Raised in case of a wrong data format.
     :throws: :obj:`tango.DevFailed`: Raised in case of a device failure.
     :throws: TypeError: Raised in case of an incorrect type of input arguments.
     __GREEN_RAISES__

     .. note::
         There are two possibilities for the attr parameter: if you give attribute name,
         then PyTango will do additional IO to fetch AttributeInfoEx object from device server,
         since we must know to which c++ data type cast python value. If you would like to avoid this IO you
         can give AttributeInfoEx instead of attribute name.

     .. versionadded:: 7.0.0

     .. versionadded:: 8.1.0
         *green_mode* parameter.
         *wait* parameter.
         *timeout* parameter.

    .. versionchanged:: 10.1.0

         - attr_name parameter was renamed to attr
         - added support for AttributeInfoEx for attr parameter

    """
    result = self._write_read_attribute(attr, value, extract_as)
    return __check_read_attribute(result)


def __DeviceProxy__write_read_attributes(
    self,
    attr_values: list[tuple[str | AttributeInfoEx, Any]],
    attr_read_names: list[str],
    extract_as: ExtractAs = ExtractAs.Numpy,
) -> DeviceAttribute:
    """

    Write then read attribute(s) in a single network call. By
    default (serialisation by device), the execution of this
    call in the server can't be interrupted by other clients.
    On the server side, attribute(s) are first written and
    if no exception has been thrown during the write phase,
    attributes will be read.

    :param attr_values: A list of pairs (attr, value). See write_attribute
    :type attr_values: sequence<(str | AttributeInfoEx, Any)>

    :param attr_read_names: A list of attributes to read.
    :type attr_read_names: sequence<str>

    :param extract_as: Defaults to numpy.
    :type extract_as: :obj:`tango.ExtractAs`

    __GREEN_KWARGS_DESCRIPTION__

    :returns: A sequence DeviceAttribute object with readout values.
    :rtype: sequence<:obj:`tango.DeviceAttribute`>

    :throws: :obj:`tango.ConnectionFailed`: Raised in case of a connection failure.
    :throws: :obj:`tango.CommunicationFailed`: Raised in case of a communication failure.
    :throws: :obj:`tango.DeviceUnlocked`: Raised in case of an unlocked device.
    :throws: :obj:`tango.WrongData`: Raised in case of a wrong data format.
    :throws: :obj:`tango.DevFailed`: Raised in case of a device failure.
    :throws: TypeError: Raised in case of an incorrect type of input arguments.
    __GREEN_RAISES__

    New in PyTango 9.2.0

    .. note::
        For each pair of values there are two possibilities for the
        attr parameter: if you give attribute name, then PyTango must
        fetch attribute info for this attribute from server by additional synchronous(!)
        IO, since we must know to which c++ data type cast each python value.
        If you would like to avoid this IO you must give AttributeInfoEx instead of
        attribute name for each(!) pair of values.

    .. versionchanged:: 10.1.0

        - name_val parameter was renamed to attr_values
        - added support for AttributeInfoEx for attr_values parameter.


    """
    return self._write_read_attributes(attr_values, attr_read_names, extract_as)


def __DeviceProxy__get_property(
    self,
    propname: (
        str
        | DbDatum
        | DbData
        | list[str | bytes | DbDatum]
        | dict[str, DbDatum]
        | dict[str, list[str]]
        | dict[str, object]
    ),
    value=None,
) -> dict[str, list[str]]:
    """

    Get a (list) property(ies) for an attribute.

    :param propname: Can be one of the following: \n
                    1. :py:obj:`str` [in] - Single property data to be fetched. \n
                    2. :py:obj:`~tango.DbDatum` [in] - Single property data to be fetched. \n
                    3. :py:obj:`~tango.DbData` [in] - Several property data to be fetched. \n
                    4. :py:obj:`list`\\[:py:obj:`str` | :py:obj:`bytes` ] [in] - Several property data to be fetched. \n
                    5. :py:obj:`list`\\[:py:obj:`tango.DbDatum`] [in] - Several property data to be fetched. \n
                    6. :py:obj:`dict`\\[:py:obj:`str`, :py:obj:`object`] [in] - Keys are property names
                       to be fetched (values are ignored). \n
                    7. :py:obj:`dict`\\[:py:obj:`str`, :obj:`tango.DbDatum`] [in] - Several `DbDatum.name` are
                       property names to be fetched (keys are ignored). \n

    :param value: Optional. For propname overloads with :py:obj:`str` and :py:obj:`list`\\[:py:obj:`str`] will be filed with the property values, if provided.
    :type value: :obj:`tango.DbData`, optional

    __GREEN_KWARGS_DESCRIPTION__

    :returns: A :obj:`dict` object, which keys are the property names the value
              associated with each key being a sequence of strings being the
              property value.

    :throws:
        :py:obj:`TypeError`: Raised in case of propname has the wrong type. \n
        :py:obj:`tango.NonDbDevice`: Raised in case of a non-database device error. \n
        :py:obj:`tango.ConnectionFailed`: Raised on connection failure with the database. \n
        :py:obj:`tango.CommunicationFailed`: Raised on communication failure with the database. \n
        :py:obj:`tango.DevFailed`: Raised on a device failure from the database device.`

    __GREEN_RAISES__

    .. versionadded:: 10.1.0: overloads with :obj:`dict` as propname parameter

    .. versionchanged:: 10.1.0: raises if propname has an invalid type instead of returning None
    """

    return get_property_from_db(self, propname, value)


def __DeviceProxy__put_property(
    self,
    value: (
        str
        | DbDatum
        | DbData
        | list[str | bytes | DbDatum]
        | dict[str, DbDatum]
        | dict[str, list[str]]
        | dict[str, object]
    ),
) -> None:
    """
    Insert or update a list of properties for this attribute.

    :param value: Can be one of the following: \n
                    1. :py:obj:`str` - Single property data to be inserted. \n
                    2. :py:obj:`~tango.DbDatum` - Single property data to be inserted. \n
                    3. :py:obj:`~tango.DbData` - Several property data to be inserted. \n
                    4. :py:obj:`list`\\[:py:obj:`str` | :py:obj:`bytes` | :py:obj:`~tango.DbDatum`] - Several property data to be inserted. \n
                    5. :py:obj:`dict`\\[:py:obj:`str`, :py:obj:`~tango.DbDatum`] -
                        DbDatum is property to be inserted (keys are ignored). \n
                    6. :py:obj:`dict`\\[:py:obj:`str`, :py:obj:`list`\\[:py:obj:`str`]] - Keys are property names,
                        and value has data to be inserted. \n
                    7. :py:obj:`dict`\\[:py:obj:`str`, :py:obj:`object`] - Keys are property names, and `str(obj)` is property value.

    __GREEN_KWARGS_DESCRIPTION__

    :throws:
        :py:obj:`TypeError`: Raised in case of value has the wrong type. \n
        :py:obj:`tango.NonDbDevice`: Raised in case of a non-database device error. \n
        :py:obj:`tango.ConnectionFailed`: Raised on connection failure with the database. \n
        :py:obj:`tango.CommunicationFailed`: Raised on communication failure with the database. \n
        :py:obj:`tango.DevFailed`: Raised on a device failure from the database device.`

    __GREEN_RAISES__
    """
    value = parameter_2_dbdata(value, "value")
    return self._put_property(value)


def __DeviceProxy__delete_property(
    self,
    value: (
        str
        | DbDatum
        | DbData
        | list[str | bytes | DbDatum]
        | dict[str, DbDatum]
        | dict[str, list[str]]
        | dict[str, object]
    ),
) -> None:
    """
    Delete a the given of properties for this attribute.
    :param value: Can be one of the following: \n
                    1. :py:obj:`str` [in] - Single property data to be deleted. \n
                    2. :py:obj:`~tango.DbDatum` [in] - Single property data to be deleted. \n
                    3. :py:obj:`~tango.DbData` [in] - Several property data to be deleted. \n
                    4. :py:obj:`list`\\[:py:obj:`str`] [in] - Several property data to be deleted. \n
                    5. :py:obj:`list`\\[:py:obj:`tango.DbDatum`] [in] - Several property data to be deleted. \n
                    6. :py:obj:`dict`\\[:py:obj:`str`, :py:obj:`object`] [in] - Keys are property names
                       to be deleted (values are ignored). \n
                    7. :py:obj:`dict`\\[:py:obj:`str`, :obj:`tango.DbDatum`] [in] - Several `DbDatum.name` are
                       property names to be deleted (keys are ignored). \n

    __GREEN_KWARGS_DESCRIPTION__

    :throws:
        :py:obj:`TypeError`: Raised in case of value has the wrong type. \n
        :py:obj:`tango.NonDbDevice`: Raised in case of a non-database device error. \n
        :py:obj:`tango.ConnectionFailed`: Raised on connection failure with the database. \n
        :py:obj:`tango.CommunicationFailed`: Raised on communication failure with the database. \n
        :py:obj:`tango.DevFailed`: Raised on a device failure from the database device.`

    __GREEN_RAISES__

    """
    new_value = parameter_2_dbdata(value, "value")
    return self._delete_property(new_value)


def __DeviceProxy__get_property_list(self, filter, array=None) -> list[str]:
    """
    get_property_list(self, filter, array=None, __GREEN_KWARGS__) -> obj

            Get the list of property names for the device. The parameter
            filter allows the user to filter the returned name list. The
            wildcard character is '*'. Only one wildcard character is
            allowed in the filter parameter.

    :param filter: The filter wildcard.
    :type filter: str

    :param array: Optional. An array to be filled with the property names. If `None`, a new list will be created internally with the values. Defaults to `None`.
    :type array: sequence obj or None, optional

    __GREEN_KWARGS_DESCRIPTION__

    :returns: The given array filled with the property names, or a new list if `array` is `None`.
    :rtype: sequence obj

    :throws: :obj:`tango.NonDbDevice`: Raised in case of a non-database device error.
    :throws: :obj:`tango.WrongNameSyntax`: Raised in case of incorrect syntax in the name.
    :throws: :obj:`tango.ConnectionFailed`: Raised in case of a connection failure with the database.
    :throws: :obj:`tango.CommunicationFailed`: Raised in case of a communication failure with the database.
    :throws: :obj:`tango.DevFailed`: Raised in case of a device failure from the database device.
    :throws: TypeError: Raised in case of an incorrect type of input arguments.
    __GREEN_RAISES__

    :versionadded:: 7.0.0
    """

    if array is None:
        new_array = StdStringVector()
        self._get_property_list(filter, new_array)
        return new_array

    if isinstance(array, StdStringVector):
        self._get_property_list(filter, array)
        return array
    elif isinstance(array, collections.abc.Sequence):
        new_array = StdStringVector()
        self._get_property_list(filter, new_array)
        StdStringVector_2_seq(new_array, array)
        return array

    raise TypeError("array must be a mutable sequence<string>")


def __DeviceProxy__get_attribute_config(
    self, value
) -> AttributeInfoEx | AttributeInfoList:
    """
    get_attribute_config(self, name, __GREEN_KWARGS__) -> AttributeInfoEx
    get_attribute_config(self, names, __GREEN_KWARGS__) -> AttributeInfoList

        Return the attribute configuration for a single or a list of attribute(s). To get all the
        attributes pass a sequence containing the constant :obj:`tango.constants.AllAttr`

        Deprecated: use get_attribute_config_ex instead

    :param name: Attribute name.
    :type name: str

    :param names: Attribute names.
    :type names: sequence<str>

    __GREEN_KWARGS_DESCRIPTION__

    :returns: An `AttributeInfoEx` or `AttributeInfoList` object containing the attribute(s) information.
    :rtype: Union[:obj:`tango.AttributeInfoEx`, :obj:`tango.AttributeInfoList`]

    :throws: :obj:`tango.ConnectionFailed`: Raised in case of a connection failure.
    :throws: :obj:`tango.CommunicationFailed`: Raised in case of a communication failure.
    :throws: :obj:`tango.DevFailed`: Raised in case of a device failure.
    :throws: TypeError: Raised in case of an incorrect type of input arguments.
    __GREEN_RAISES__
    """
    if isinstance(value, StdStringVector) or is_pure_str(value):
        return self._get_attribute_config(value)
    elif isinstance(value, collections.abc.Sequence):
        v = seq_2_StdStringVector(value)
        return self._get_attribute_config(v)

    raise TypeError("value must be a string or a sequence<string>")


def __DeviceProxy__get_attribute_config_ex(self, value) -> AttributeInfoListEx:
    """
    get_attribute_config_ex(self, name or sequence(names), __GREEN_KWARGS__) -> AttributeInfoListEx :

        Return the extended attribute configuration for a single attribute or for the list of
        specified attributes. To get all the attributes pass a sequence
        containing the constant tango.constants.AllAttr.

    :param name: Attribute name or attribute names. Can be a single string (for one attribute) or a sequence of strings (for multiple attributes).
    :type name: str or sequence(str)

    __GREEN_KWARGS_DESCRIPTION__

    :returns: An `AttributeInfoListEx` object containing the attribute information.
    :rtype: :obj:`tango.AttributeInfoListEx`

    :throws: :obj:`tango.ConnectionFailed`: Raised in case of a connection failure.
    :throws: :obj:`tango.CommunicationFailed`: Raised in case of a communication failure.
    :throws: :obj:`tango.DevFailed`: Raised in case of a device failure.
    __GREEN_RAISES__
    """
    if isinstance(value, StdStringVector):
        return self._get_attribute_config_ex(value)
    elif is_pure_str(value):
        v = StdStringVector()
        v.append(value)
        return self._get_attribute_config_ex(v)
    elif isinstance(value, collections.abc.Sequence):
        v = seq_2_StdStringVector(value)
        return self._get_attribute_config_ex(v)

    raise TypeError("value must be a string or a sequence<string>")


def __DeviceProxy__get_command_config(
    self, value=(constants.AllCmd,)
) -> CommandInfoList | CommandInfo:
    """
    get_command_config(self, __GREEN_KWARGS__) -> CommandInfoList
    get_command_config(self, name, __GREEN_KWARGS__) -> CommandInfo
    get_command_config(self, names, __GREEN_KWARGS__) -> CommandInfoList

        Return the command configuration for single/list/all command(s).

    :param name: Command name. Used when querying information for a single command.
    :type name: str, optional

    :param names: Command names. Used when querying information for multiple commands. This parameter should not be used simultaneously with 'name'.
    :type names: sequence<str>, optional

    __GREEN_KWARGS_DESCRIPTION__

    :returns: A `CommandInfoList` object containing the commands information if multiple command names are provided, or a `CommandInfo` object if a single command name is provided.
    :rtype: :obj:`tango.CommandInfoList` or :obj:`tango.CommandInfo`

    :throws: :obj:`tango.ConnectionFailed`: Raised in case of a connection failure.
    :throws: :obj:`tango.CommunicationFailed`: Raised in case of a communication failure.
    :throws: :obj:`tango.DevFailed`: Raised in case of a device failure.
    :throws: TypeError: Raised in case of an incorrect type of input arguments.
    __GREEN_RAISES__
    """
    if isinstance(value, StdStringVector) or is_pure_str(value):
        return self._get_command_config(value)
    elif isinstance(value, collections.abc.Sequence):
        v = seq_2_StdStringVector(value)
        return self._get_command_config(v)

    raise TypeError("value must be a string or a sequence<string>")


def __DeviceProxy__set_attribute_config(self, value) -> None:
    """
    set_attribute_config(self, attr_info, __GREEN_KWARGS__) -> None
    set_attribute_config(self, attr_info_ex, __GREEN_KWARGS__) -> None

        Change the attribute configuration/extended attribute configuration for the specified attribute(s)

    :param attr_info: Attribute information. This parameter is used when providing basic attribute(s) information.
    :type attr_info: Union[:obj:`tango.AttributeInfo`, Sequence[:obj:`tango.AttributeInfo`]], optional

    :param attr_info_ex: Extended attribute information. This parameter is used when providing extended attribute information. It should not be used simultaneously with 'attr_info'.
    :type attr_info_ex: Union[:obj:`tango.AttributeInfoEx`, Sequence[:obj:`tango.AttributeInfoEx`]], optional

    __GREEN_KWARGS_DESCRIPTION__

    :returns: None

    :throws: :obj:`tango.ConnectionFailed`: Raised in case of a connection failure.
    :throws: :obj:`tango.CommunicationFailed`: Raised in case of a communication failure.
    :throws: :obj:`tango.DevFailed`: Raised in case of a device failure.
    :throws: TypeError: Raised in case of an incorrect type of input arguments.
    __GREEN_RAISES__

    """
    if isinstance(value, AttributeInfoEx):
        v = AttributeInfoListEx()
        v.append(value)
    elif isinstance(value, AttributeInfo):
        v = AttributeInfoList()
        v.append(value)
    elif isinstance(value, AttributeInfoList):
        v = value
    elif isinstance(value, AttributeInfoListEx):
        v = value
    elif isinstance(value, collections.abc.Sequence):
        if not len(value):
            return
        if isinstance(value[0], AttributeInfoEx):
            v = AttributeInfoListEx()
        elif isinstance(value[0], AttributeInfo):
            v = AttributeInfoList()
        else:
            raise TypeError(
                "Value must be a AttributeInfo, AttributeInfoEx, "
                "sequence<AttributeInfo> or sequence<AttributeInfoEx"
            )
        for i in value:
            v.append(i)
    else:
        raise TypeError(
            "Value must be a AttributeInfo, AttributeInfoEx, "
            "sequence<AttributeInfo> or sequence<AttributeInfoEx"
        )

    return self._set_attribute_config(v)


def __DeviceProxy__get_event_map_lock(self):
    """
    Internal helper method"""
    if "_subscribed_events_lock" not in self.__dict__:
        # do it like this instead of self._subscribed_events = dict() to avoid
        # calling __setattr__ which requests list of tango attributes from device
        self.__dict__["_subscribed_events_lock"] = threading.Lock()
    return self._subscribed_events_lock


def __DeviceProxy__get_event_map(self):
    """
    Internal helper method"""
    if "_subscribed_events" not in self.__dict__:
        # do it like this instead of self._subscribed_events = dict() to avoid
        # calling __setattr__ which requests list of tango attributes from device
        self.__dict__["_subscribed_events"] = dict()
    return self._subscribed_events


def __DeviceProxy__subscribe_event(self, *args, **kwargs) -> int:
    """
    subscribe_event(self, attr_name, event_type, cb, sub_mode=EventSubMode.SyncRead, extract_as=ExtractAs.Numpy, *, __GREEN_KWARGS__) -> int
    subscribe_event(self, attr_name, event_type, queuesize, sub_mode=EventSubMode.SyncRead, extract_as=ExtractAs.Numpy, *, __GREEN_KWARGS__) -> int
    subscribe_event(self, event_type, cb, sub_mode=EventSubMode.SyncRead, *, __GREEN_KWARGS__) -> int

    .. note::
        This function is heavily overloaded, and includes three additional signatures that
        have been deprecated (see warning below).

    The client call to subscribe for event reception.  There are two main categories:

    * Subscribe to events for a specific attribute, like change events, providing
      either:

      * a callback function for immediate notification (*push* callback model), or
      * a queue length, allowing events to be processed later (*pull* callback model).

    * Subscribe to device-level events (not linked to a specific attribute), like
      interface-changed events. These require a callback function (*push* callback model).

    More details of the *push* and *pull* callback models are provided in the
    :external+tangodoc:ref:`cppTango Events API docs <events-tangoclient-api>`.

    :param attr_name: The device attribute name which will be sent as an event, e.g., "current".
    :type attr_name: str

    :param event_type: The event reason, which must be one of the enumerated values in `EventType`. This includes:

        * `EventType.CHANGE_EVENT`
        * `EventType.ALARM_EVENT`
        * `EventType.PERIODIC_EVENT`
        * `EventType.ARCHIVE_EVENT`
        * `EventType.USER_EVENT`
        * `EventType.DATA_READY_EVENT`
        * `EventType.ATTR_CONF_EVENT`
        * `EventType.INTERFACE_CHANGE_EVENT`
    :type event_type: :obj:`tango.EventType`

    :param cb:
        Any callable object or an object with a callable ``push_event`` method
        (i.e., use the *push* callback model).
        The callable has the signature ``def cb(event)`` (or ``async def cb(event)``
        for asyncio green mode DeviceProxy objects).  The ``event`` parameter's data
        type depends on the type of event subscribed to. In most cases it is
        :obj:`tango.EventData`.  Special cases are `EventType.DATA_READY`,
        `EventType.ATTR_CONF_EVENT`, and `EventType.INTERFACE_CHANGE_EVENT` -
        see :ref:`event-arrived-structures`.

    :type cb: callable

    :param queuesize:
        The size of the event reception buffer (i.e., use the *pull* callback model).
        The event reception buffer is implemented as a round-robin buffer.
        This way the client can set up different ways to receive events:

        * Event reception buffer size = 1 : The client is interested only in the value of the last event received. All other events that have been received since the last reading are discarded.
        * Event reception buffer size > 1 : The client has chosen to keep an event history of a given size. When more events arrive since the last reading, older events will be discarded.
        * Event reception buffer size = tango.constants.ALL_EVENTS : The client buffers all received events. The buffer size is unlimited and only restricted by the available memory for the client.
    :type queuesize: int

    :param sub_mode: The event subscription mode.
    :type sub_mode: :obj:`tango.EventSubMode`

    :param extract_as: In which format to return the attribute values. Default: ExtractAs.NumPy
    :type extract_as: ExtractAs

    __GREEN_KWARGS_DESCRIPTION__

    :returns: An event id which has to be specified when unsubscribing from this event.
    :rtype: int

    :throws: :obj:`tango.EventSystemFailed`: Raised in case of a failure in the event system.
    :throws: :obj:`tango.DevFailed`: Raised in case of general communication failures.
    :throws: :obj:`TypeError`: Raised in case of an incorrect type of input arguments.
    __GREEN_RAISES__

    .. deprecated:: PyTango 10.1.0

        The following signatures, which use the ``filters`` and/or ``stateless``
        parameters, are deprecated. The version for removal has not been decided, but
        the earliest is version 11.0.0.

        * *subscribe_event(self, attr_name, event_type, cb, filters=[], stateless=False, extract_as=ExtractAs.Numpy, __GREEN_KWARGS__) -> int*
        * *subscribe_event(self, attr_name, event_type, queuesize, filters=[], stateless=False, extract_as=ExtractAs.Numpy, __GREEN_KWARGS__) -> int*
        * *subscribe_event(self, event_type, cb, stateless=False, *, __GREEN_KWARGS__) -> int*

        ``filters`` (*sequence<str>*)
            The filters apply to the original Notifd-based event system, not the
            current ZeroMQ-based event system (added in Tango 8). Notifd support is
            scheduled for removal in Tango 11.

            A variable length list of name-value pairs which define additional filters
            for events. Provide an empty list, if this feature is not needed
            (typically, for all Tango servers from version 8).

            Filters cannot be used with the `sub_mode` parameter.

        ``stateless`` (*bool*)
            Instead of setting ``stateless=True`` use ``sub_mode=EventSubMode.Stateless``.

            When this flag is set to false, an exception will be thrown if the event
            subscription encounters a problem. With the stateless flag set to true,
            the event subscription will not raise an exception, even if the corresponding
            device is not running, or the attribute doesn't exist.
            A keep-alive thread will attempt to subscribe for the specified event
            every 10 seconds, executing a callback with the corresponding exception at
            every retry.

            The ``stateless`` flag cannot be used with the `sub_mode` parameter.

    .. versionadded:: 10.1.0
        Three new signatures using the *sub_mode* parameter.

    .. versionchanged:: 10.1.0
        All parameters can now be passed as keyword arguments.
    """
    _check_only_allowed_kwargs(
        kwargs,
        {
            "attr_name",
            "event_type",
            "queuesize",
            "sub_mode",
            "extract_as",
            "event_type",
            "cb",
            "filters",
            "stateless",
            "green_mode",
        },
    )

    if (args and isinstance(args[0], EventType)) or kwargs.get(
        "event_type", None
    ) == EventType.INTERFACE_CHANGE_EVENT:
        return __DeviceProxy__subscribe_event_global(self, *args, **kwargs)
    # The first argument is the attribute name
    else:
        return __DeviceProxy__subscribe_event_attrib(self, *args, **kwargs)


def __DeviceProxy__subscribe_event_global(self, *args, **kwargs):
    """
    Various signatures to resolve:
    index:                0             1     2
    subscribe_event(self, event_type,   cb,   stateless=False,                  *, green_mode=None) -> int
    subscribe_event(self, event_type,   cb,   sub_mode=EventSubMode.SyncRead,   *, green_mode=None) -> int
    """

    event_type, _ = _get_and_check_param(
        args,
        kwargs,
        "event_type",
        0,
        type_check=lambda x: isinstance(x, EventType),
    )

    if event_type != EventType.INTERFACE_CHANGE_EVENT:
        raise TypeError("This method is only for Interface Change Events")

    user_cb, _ = _get_and_check_param(
        args,
        kwargs,
        "cb",
        1,
        type_check=_is_callable_for_event,
        type_error_message="Parameter 'cb' must be callable object or "
        "an object with a 'push_event' method.",
    )

    sub_mode = EventSubMode.SyncRead
    stateless = None

    try:
        sub_mode, _ = _get_and_check_param(
            args,
            kwargs,
            "sub_mode",
            2,
            type_check=lambda x: isinstance(x, EventSubMode),
        )
    except (IndexError, PositionalArgTypeError):
        try:
            stateless, _ = _get_and_check_param(
                args,
                kwargs,
                "stateless",
                2,
                type_check=lambda x: isinstance(x, bool),
                type_error_message=" must be either EventSubMode (sub_mode) or bool (stateless) type.",
            )
        except IndexError:
            pass

    green_mode = kwargs.get("green_mode", None)

    cbfn = __EventCallBack()
    cbfn.push_event = green_callback(
        _callable_for_event(user_cb), obj=self, green_mode=green_mode
    )

    if stateless is not None:
        warnings.warn(
            "The 'stateless' parameter is deprecated. "
            "The version for removal has not been decided, but the earliest is version 11.0.0. "
            "Please, update calls to use the 'sub_mode' parameter instead.",
            category=DeprecationWarning,
        )
        event_id = self.__subscribe_event_global_with_stateless_flag(
            event_type, cbfn, stateless
        )
    else:
        event_id = self.__subscribe_event_global_with_sub_mode(
            event_type, cbfn, sub_mode
        )

    _add_event_to_map(self, event_id, cbfn, event_type, "dummy")
    return event_id


def __DeviceProxy__subscribe_event_attrib(self, *args, **kwargs):
    """
    Various signatures to resolve:
              index:          0          1           2          3                               4
        subscribe_event(self, attr_name, event_type, cb,        sub_mode=EventSubMode.SyncRead, extract_as=ExtractAs.Numpy,                             *,  green_mode=None) -> int
        subscribe_event(self, attr_name, event_type, queuesize, sub_mode=EventSubMode.SyncRead, extract_as=ExtractAs.Numpy,                             *,  green_mode=None) -> int

    DEPRECATED, but still valid:
              index:          0          1           2          3                               4                           5
        subscribe_event(self, attr_name, event_type, cb,        filters=[],                     stateless=False,            extract_as=ExtractAs.Numpy, *, green_mode=None) -> int
        subscribe_event(self, attr_name, event_type, queuesize, filters=[],                     stateless=False,            extract_as=ExtractAs.Numpy, *, green_mode=None) -> int

    """

    attr_name, _ = _get_and_check_param(
        args,
        kwargs,
        "attr_name",
        0,
        type_check=is_pure_str,
    )

    event_type, _ = _get_and_check_param(
        args,
        kwargs,
        "event_type",
        1,
        type_check=lambda x: isinstance(x, EventType),
    )

    if "cb" in kwargs and "queuesize" in kwargs:
        raise TypeError(
            f"Parameters 'cb' and 'queuesize' cannot be used together'. {args=}, {kwargs=}"
        )
    user_cb = None
    queuesize = None
    try:
        user_cb, _ = _get_and_check_param(
            args,
            kwargs,
            "cb",
            2,
            type_check=_is_callable_for_event,
            type_error_message="Parameter 'cb' must be callable object or "
            "an object with a 'push_event' method",
        )
    except (IndexError, PositionalArgTypeError):
        try:
            queuesize, _ = _get_and_check_param(
                args, kwargs, "queuesize", 2, type_check=is_integer
            )
        except (IndexError, PositionalArgTypeError):
            raise TypeError(
                f"Either parameter 'queuesize' must be an integer, or parameter 'cb' "
                f"must be callable object or an object with a 'push_event' method. "
                f"{args=}, {kwargs=}."
            )

    if "sub_mode" in kwargs and "filters" in kwargs:
        raise TypeError(
            f"Parameters 'sub_mode' and 'filters' cannot be used together. {args=}, {kwargs=}"
        )
    if "sub_mode" in kwargs and "stateless" in kwargs:
        raise TypeError(
            f"Parameters 'sub_mode' and 'stateless' cannot be used together. {args=}, {kwargs=}"
        )
    sub_mode = None
    filters = None
    sub_mode_from_caller = False
    filters_from_caller = False
    arg_shift = 0
    try:
        sub_mode, _ = _get_and_check_param(
            args,
            kwargs,
            "sub_mode",
            3,
            type_check=lambda x: isinstance(x, EventSubMode),
            type_error_message="Parameter 'sub_mode' must be of type EventSubMode",
        )
        sub_mode_from_caller = True
    except (IndexError, PositionalArgTypeError) as sub_mode_exc:
        try:
            filters, _ = _get_and_check_param(
                args,
                kwargs,
                "filters",
                3,
                type_check=is_non_str_seq,
                type_error_message="Parameter 'filters' must be sequence of str",
            )
            filters_from_caller = True
            if "filters" not in kwargs:
                arg_shift = 1
        except (IndexError, PositionalArgTypeError) as filters_exc:
            if isinstance(sub_mode_exc, PositionalArgTypeError):
                raise sub_mode_exc
            elif isinstance(filters_exc, PositionalArgTypeError):
                raise filters_exc
            # fallback to defaults
            sub_mode = EventSubMode.SyncRead
            filters = []

    stateless = None
    stateless_from_caller = False
    if not sub_mode_from_caller:
        stateless, used_default = _get_and_check_param(
            args,
            kwargs,
            "stateless",
            4,
            default=False,
            type_check=lambda x: isinstance(x, bool),
            type_error_message="Parameter 'stateless' must be of type bool",
        )
        stateless_from_caller = not used_default

    extract_as, _ = _get_and_check_param(
        args,
        kwargs,
        "extract_as",
        4 + arg_shift,
        type_check=lambda x: isinstance(x, ExtractAs),
        default=ExtractAs.Numpy,
        type_error_message="Parameter 'extract_as' must be of type ExtractAs",
    )

    green_mode = kwargs.get("green_mode", None)

    # at least one of "cb" or "queuesize" must be provided
    if user_cb:
        cb_or_queuesize = __EventCallBack()
        cb_or_queuesize.push_event = green_callback(
            _callable_for_event(user_cb), obj=self, green_mode=green_mode
        )
    elif queuesize is not None:
        cb_or_queuesize = queuesize
    else:
        raise TypeError(
            f"Either parameter 'queuesize' must be an integer, or parameter 'cb' "
            f"must be callable object or an object with a 'push_event' method. "
            f"{args=}, {kwargs=}."
        )

    if sub_mode_from_caller:
        use_sub_mode = True
    elif filters_from_caller or stateless_from_caller:
        use_sub_mode = False
    else:
        use_sub_mode = True

    if use_sub_mode:
        event_id = self.__subscribe_event_attrib_with_sub_mode(
            attr_name, event_type, cb_or_queuesize, sub_mode, extract_as
        )
    else:
        warnings.warn(
            "The 'stateless' and 'filters' parameters are deprecated. "
            "The version for removal has not been decided, but the earliest is version 11.0.0. "
            "Please, update calls to use the 'sub_mode' parameter instead.",
            category=DeprecationWarning,
        )
        event_id = self.__subscribe_event_attrib_with_stateless_flag(
            attr_name, event_type, cb_or_queuesize, stateless, extract_as, filters
        )

    _add_event_to_map(self, event_id, cb_or_queuesize, event_type, attr_name)

    return event_id


class _DummyValue:
    pass


class PositionalArgTypeError(TypeError):
    pass


def _get_and_check_param(
    user_args: Tuple[Any, ...],
    user_kwargs: Dict[str, Any],
    name: str,
    index: int,
    default: Any = _DummyValue(),
    type_check: Any = None,
    type_error_message: str = "",
) -> Tuple[Any, bool]:
    used_default = False
    if name in user_kwargs:
        value = user_kwargs[name]
        if callable(type_check) and not type_check(value):
            raise TypeError(
                f"Invalid type for parameter '{name}={value}': {type_error_message}"
            )
        return value, used_default

    if len(user_args) > index:
        value = user_args[index]
    else:
        if isinstance(default, _DummyValue):
            raise IndexError(
                f"Expected parameter '{name}' as either positional arg {index + 1}, "
                f"(but only {len(user_args)} provided), or as keyword argument. "
                f"{user_args=}, {user_kwargs=}."
            )
        value = default
        used_default = True

    if callable(type_check) and not type_check(value):
        raise PositionalArgTypeError(
            f"Invalid type for parameter '{name}' at position {index}: "
            f"{type_error_message}. Received: {type(value)}"
        )

    return value, used_default


def _callable_for_event(fn):
    if isinstance(fn, collections.abc.Callable):
        return fn
    elif hasattr(fn, "push_event") and isinstance(
        fn.push_event, collections.abc.Callable
    ):
        return fn.push_event
    else:
        return None


def _is_callable_for_event(fn):
    return _callable_for_event(fn) is not None


def _add_event_to_map(self, event_id, cb_or_queuesize, event_type, attr_name):
    with self.__get_event_map_lock():
        se = self.__get_event_map()
        evt_data = se.get(event_id)
        if evt_data is None:
            se[event_id] = (cb_or_queuesize, event_type, attr_name)
            return

        # Raise exception
        desc = textwrap.dedent(
            f"""\
            Internal PyTango error:
            {self}.subscribe_event({attr_name}, {event_type}) already has key {event_id} assigned to ({evt_data[2]}, {evt_data[1]})
            Please report error to PyTango"""
        )
        Except.throw_exception("Py_InternalError", desc, "DeviceProxy.subscribe_event")


def __dummy_event_receiver(_event):
    pass


async def __async_dummy_event_receiver(_event):
    pass


def __DeviceProxy__unsubscribe_event(self, event_id) -> None:
    """
    unsubscribe_event(self, event_id, __GREEN_KWARGS__) -> None

        Unsubscribes a client from receiving the event specified by event_id.

    :param event_id: The event identifier returned by `DeviceProxy::subscribe_event()`. Unlike in TangoC++, this implementation checks that the `event_id` has been subscribed to in this `DeviceProxy`.
    :type event_id: int

    __GREEN_KWARGS_DESCRIPTION__

    :returns: None

    :throws: :obj:`tango.EventSystemFailed`: Raised in case of a failure in the event system.
    :throws: KeyError: Raised if the specified `event_id` is not found or not subscribed in this `DeviceProxy`.
    __GREEN_RAISES__
    """
    events_del = set()
    timestamp = time.time()
    se = self.__get_event_map()

    with self.__get_event_map_lock():
        # first delete event callbacks that have expired
        for evt_id, (_, expire_time) in self._pending_unsubscribe.items():
            if expire_time <= timestamp:
                events_del.add(evt_id)
        for evt_id in events_del:
            del self._pending_unsubscribe[evt_id]

        # unsubscribe and put the callback in the pending unsubscribe callbacks
        try:
            evt_info = se[event_id]
        except KeyError:
            raise KeyError(
                "This device proxy does not own this subscription " + str(event_id)
            )

        del se[event_id]

        # if we subscribe callback, we have to keep it alive for sometime,
        # in case there is delayed event comming
        if hasattr(evt_info[0], "push_event"):
            if self.get_green_mode() == GreenMode.Asyncio:
                evt_info[0].push_event = __async_dummy_event_receiver
            else:
                evt_info[0].push_event = __dummy_event_receiver

            self._pending_unsubscribe[event_id] = (
                evt_info[0],
                timestamp + _UNSUBSCRIBE_LIFETIME,
            )

    self.__unsubscribe_event(event_id)


def __DeviceProxy__unsubscribe_event_all(self):
    with self.__get_event_map_lock():
        se = self.__get_event_map()
        event_ids = list(se.keys())
        se.clear()
    for event_id in event_ids:
        self.__unsubscribe_event(event_id)


def __DeviceProxy__get_events(
    self, event_id, callback=None, extract_as=ExtractAs.Numpy
):
    """
    get_events(self, event_id, callback=None, extract_as=Numpy) -> None

        The method extracts all waiting events from the event reception buffer.

        If callback is not None, it is executed for every event. During event
        subscription the client must have chosen the pull model for this event.
        The callback will receive a parameter of type EventData,
        AttrConfEventData or DataReadyEventData depending on the type of the
        event (event_type parameter of subscribe_event).

        If callback is None, the method extracts all waiting events from the
        event reception buffer. The returned event_list is a vector of
        EventData, AttrConfEventData or DataReadyEventData pointers, just
        the same data the callback would have received.

    :param event_id: The event identifier returned by the `DeviceProxy.subscribe_event()` method.
    :type event_id: int
    :param callback: Any callable object or any object with a "push_event" method.
    :type callback: callable
    :param extract_as: (Description Needed)
    :type extract_as: :obj:`tango.ExtractAs`

    :returns: None

    :throws: :obj:`tango.EventSystemFailed`: Raised in case of a failure in the event system.
    :throws: TypeError: Raised in case of an incorrect type of input arguments.
    :throws: ValueError: Raised in case of an invalid value.

    :see also: :meth:`~tango.DeviceProxy.subscribe_event`

    """
    if callback is None:
        queuesize, event_type, attr_name = self.__get_event_map().get(
            event_id, (None, None, None)
        )
        if event_type is None:
            raise ValueError(
                f"Invalid event_id. You are not subscribed to event {str(event_id)}."
            )
        if event_type in (
            EventType.CHANGE_EVENT,
            EventType.ALARM_EVENT,
            EventType.PERIODIC_EVENT,
            EventType.ARCHIVE_EVENT,
            EventType.USER_EVENT,
        ):
            return self.__get_data_events(event_id, extract_as)
        elif event_type in (EventType.ATTR_CONF_EVENT,):
            return self.__get_attr_conf_events(event_id)
        elif event_type in (EventType.DATA_READY_EVENT,):
            return self.__get_data_ready_events(event_id)
        elif event_type in (EventType.INTERFACE_CHANGE_EVENT,):
            return self.__get_devintr_change_events(event_id, extract_as)
        else:
            raise ValueError("Unknown event_type: " + str(event_type))
    elif isinstance(callback, collections.abc.Callable):
        cb = __EventCallBack()
        cb.push_event = callback
        return self.__get_callback_events(event_id, cb, extract_as)
    elif hasattr(callback, "push_event") and isinstance(
        callback.push_event, collections.abc.Callable
    ):
        cb = __EventCallBack()
        cb.push_event = callback.push_event
        return self.__get_callback_events(event_id, cb, extract_as)
    else:
        raise TypeError(
            "Parameter 'callback' should be None, a callable object or an object with a 'push_event' method."
        )


def __DeviceProxy___get_info_(self):
    """Protected method that gets device info once and stores it in cache"""
    if "_dev_info" not in self.__dict__:
        try:
            info = self.info()
            info_without_cyclic_reference = __TangoInfo.from_copy(info)
            self.__dict__["_dev_info"] = info_without_cyclic_reference
        except Exception:
            return __TangoInfo.from_defaults()
    return self._dev_info


def __DeviceProxy__str(self):
    is_safe_to_access_proxy = self.__dict__.get("_initialized", False)
    if is_safe_to_access_proxy:
        info = self._get_info_()
        dev_class = info.dev_class
        dev_name = self.dev_name()
    else:
        dev_class = "DeviceProxy<Unknown Device>"
        dev_name = "<object was not fully initialized>"
    return f"{dev_class}({dev_name})"


def __DeviceProxy__read_attributes(
    self, attr_names, extract_as=ExtractAs.Numpy
) -> list[DeviceAttribute]:
    """
    read_attributes(self, attr_names, extract_as=ExtractAs.Numpy, __GREEN_KWARGS__) -> sequence<DeviceAttribute>

        Read the list of specified attributes.

    :param attr_names: A list of attributes to read.
    :type attr_names: sequence<str>

    :param extract_as: In which format to return the attribute values.
    :type extract_as: :obj:`tango.ExtractAs`

    __GREEN_KWARGS_DESCRIPTION__

    :returns: A list of DeviceAttribute objects.
    :rtype: sequence<:obj:`tango.DeviceAttribute`>

    :throws: :obj:`tango.ConnectionFailed`, :obj:`tango.CommunicationFailed`, :obj:`tango.DeviceUnlocked`:
    :throws: :obj:`tango.DevFailed`: from device
    __GREEN_RAISES__

    .. versionadded:: 8.1.0
        *green_mode* parameter.
        *wait* parameter.
        *timeout* parameter.
    """
    return self._read_attributes(attr_names, extract_as)


def __DeviceProxy__write_attribute(
    self, attr: str | AttributeInfoEx, value: Any
) -> None:
    """
    write_attribute(self, attr, value, __GREEN_KWARGS__) -> None

    Write a single attribute.

    :param attr: The name or AttributeInfoEx structure of the attribute to write.
    :type attr: str | :obj:`tango.AttributeInfoEx`

    :param value: The value to write to the attribute.
    :type value: Any

    __GREEN_KWARGS_DESCRIPTION__

    :throws: :obj:`tango.ConnectionFailed`, :obj:`tango.CommunicationFailed`, :obj:`tango.DeviceUnlocked`:
    :throws: :obj:`tango.DevFailed`: from device
    __GREEN_RAISES__

    .. versionadded:: 8.1.0
        *green_mode* parameter.
        *wait* parameter.
        *timeout* parameter.

    .. versionchanged:: 10.1.0 attr_name parameter was renamed to attr

    """
    return self._write_attribute(attr, value)


def __DeviceProxy__write_attributes(self, attr_values) -> None:
    """
    write_attributes(self, name_val, __GREEN_KWARGS__) -> None

        Write the specified attributes.

    :param attr_values: A list of pairs (attr, value). See write_attribute
    :type attr_values: Sequence<(str | AttributeInfoEx, Any)>

    __GREEN_KWARGS_DESCRIPTION__

    :throws: :obj:`tango.ConnectionFailed`, :obj:`tango.CommunicationFailed`, :obj:`tango.DeviceUnlocked`:
    :throws: :obj:`tango.DevFailed`, :obj:`tango.NamedDevFailedList`: from device
    __GREEN_RAISES__

    .. versionadded:: 8.1.0
        *green_mode* parameter.
        *wait* parameter.
        *timeout* parameter.

    .. versionchanged:: 10.1.0

        - name_val parameter was renamed to attr_values
        - added support for AttributeInfoEx for attr_values parameter
    """
    return self._write_attributes(attr_values)


def __DeviceProxy__ping(self) -> int:
    """
    ping(self, __GREEN_KWARGS__) -> DevState

        A method which sends a ping to the device

    __GREEN_KWARGS_DESCRIPTION__

    :returns: time elapsed in microseconds
    :rtype: int

    __GREEN_RAISES__

    """
    return self._ping()


def __DeviceProxy__state(self) -> DevState:
    """
    state(self, __GREEN_KWARGS__) -> DevState

        A method which returns the state of the device.

    __GREEN_KWARGS_DESCRIPTION__

    :returns: A `DevState` constant.
    :rtype: DevState

    __GREEN_RAISES__

    """
    return self._state()


def __DeviceProxy__status(self) -> str:
    """
    status(self, __GREEN_KWARGS__) -> str

        A method which returns the status of the device as a string.

    __GREEN_KWARGS_DESCRIPTION__

    :returns: string describing the device status
    :rtype: str

    __GREEN_RAISES__

    """
    return self._status()


def __safe_call(fn):

    already_wrapped = hasattr(fn, "__access_wrapped__")
    if already_wrapped:
        return fn

    @functools.wraps(fn)
    def safe_call_wrapper(self, *args, **kwargs):
        if not self._initialized:
            raise RuntimeError("DeviceProxy object was not fully initialized.")
        return fn(self, *args, **kwargs)

    fn.__access_wrapped__ = True

    return safe_call_wrapper


def device_proxy_init():
    DeviceProxy.__init_orig__ = DeviceProxy.__init__
    DeviceProxy.__init__ = _trace_client(__DeviceProxy____init__)

    DeviceProxy.get_green_mode = __DeviceProxy__get_green_mode
    DeviceProxy.set_green_mode = __DeviceProxy__set_green_mode

    DeviceProxy.freeze_dynamic_interface = __DeviceProxy__freeze_dynamic_interface
    DeviceProxy.unfreeze_dynamic_interface = __DeviceProxy__unfreeze_dynamic_interface
    DeviceProxy.is_dynamic_interface_frozen = __DeviceProxy__is_dynamic_interface_frozen

    DeviceProxy.__get_cmd_cache = __DeviceProxy__get_cmd_cache
    DeviceProxy.__get_attr_cache = __DeviceProxy__get_attr_cache
    DeviceProxy.__refresh_cmd_cache = __DeviceProxy__refresh_cmd_cache
    DeviceProxy.__refresh_attr_cache = __DeviceProxy__refresh_attr_cache

    DeviceProxy.ping = green(
        _trace_client(__DeviceProxy__ping), update_signature_and_docstring=True
    )
    DeviceProxy.state = green(
        _trace_client(__DeviceProxy__state), update_signature_and_docstring=True
    )
    DeviceProxy.status = green(
        _trace_client(__DeviceProxy__status), update_signature_and_docstring=True
    )
    DeviceProxy.state = green(
        _trace_client(__DeviceProxy__state), update_signature_and_docstring=True
    )
    DeviceProxy.status = green(
        _trace_client(__DeviceProxy__status), update_signature_and_docstring=True
    )

    DeviceProxy.read_attribute = green(
        _trace_client(__DeviceProxy__read_attribute),
        update_signature_and_docstring=True,
    )
    DeviceProxy.read_attributes = green(
        _trace_client(__DeviceProxy__read_attributes),
        update_signature_and_docstring=True,
    )
    DeviceProxy.write_attribute = green(
        _trace_client(__DeviceProxy__write_attribute),
        update_signature_and_docstring=True,
    )
    DeviceProxy.write_attributes = green(
        _trace_client(__DeviceProxy__write_attributes),
        update_signature_and_docstring=True,
    )
    DeviceProxy.write_attributes = green(
        _trace_client(__DeviceProxy__write_attributes),
        update_signature_and_docstring=True,
    )
    DeviceProxy.write_read_attribute = green(
        _trace_client(__DeviceProxy__write_read_attribute),
        update_signature_and_docstring=True,
    )
    DeviceProxy.write_read_attributes = green(
        _trace_client(__DeviceProxy__write_read_attributes),
        update_signature_and_docstring=True,
    )

    DeviceProxy.read_attributes_asynch = green(
        _trace_client(__DeviceProxy__read_attributes_asynch),
        update_signature_and_docstring=True,
    )
    DeviceProxy.read_attribute_asynch = green(
        _trace_client(__DeviceProxy__read_attribute_asynch),
        update_signature_and_docstring=True,
    )

    DeviceProxy.read_attributes_reply = green(
        _trace_client(__DeviceProxy__read_attributes_reply),
        update_signature_and_docstring=True,
    )
    DeviceProxy.read_attribute_reply = green(
        _trace_client(__DeviceProxy__read_attribute_reply),
        update_signature_and_docstring=True,
    )

    DeviceProxy.write_attributes_asynch = green(
        _trace_client(__DeviceProxy__write_attributes_asynch),
        update_signature_and_docstring=True,
    )
    DeviceProxy.write_attribute_asynch = green(
        _trace_client(__DeviceProxy__write_attribute_asynch),
        update_signature_and_docstring=True,
    )

    DeviceProxy.write_attributes_reply = green(
        _trace_client(__DeviceProxy__write_attributes_reply),
        update_signature_and_docstring=True,
    )
    DeviceProxy.write_attribute_reply = green(
        _trace_client(__DeviceProxy__write_attribute_reply),
        update_signature_and_docstring=True,
    )

    DeviceProxy.get_property = green(
        _trace_client(__DeviceProxy__get_property), update_signature_and_docstring=True
    )
    DeviceProxy.put_property = green(
        _trace_client(__DeviceProxy__put_property), update_signature_and_docstring=True
    )
    DeviceProxy.delete_property = green(
        _trace_client(__DeviceProxy__delete_property),
        update_signature_and_docstring=True,
    )
    DeviceProxy.get_property_list = green(
        _trace_client(__DeviceProxy__get_property_list),
        update_signature_and_docstring=True,
    )
    DeviceProxy.get_attribute_config = green(
        _trace_client(__DeviceProxy__get_attribute_config),
        update_signature_and_docstring=True,
    )
    DeviceProxy.get_attribute_config_ex = green(
        _trace_client(__DeviceProxy__get_attribute_config_ex),
        update_signature_and_docstring=True,
    )
    DeviceProxy.set_attribute_config = green(
        _trace_client(__DeviceProxy__set_attribute_config),
        update_signature_and_docstring=True,
    )

    DeviceProxy.get_command_config = green(
        _trace_client(__DeviceProxy__get_command_config),
        update_signature_and_docstring=True,
    )

    DeviceProxy.__get_event_map = __DeviceProxy__get_event_map
    DeviceProxy.__get_event_map_lock = __DeviceProxy__get_event_map_lock

    DeviceProxy.subscribe_event = green(
        _trace_client(__DeviceProxy__subscribe_event),
        consume_green_mode=False,
        update_signature_and_docstring=True,
    )
    DeviceProxy.unsubscribe_event = green(
        _trace_client(__DeviceProxy__unsubscribe_event),
        update_signature_and_docstring=True,
    )
    DeviceProxy.get_events = _trace_client(__DeviceProxy__get_events)
    DeviceProxy.__unsubscribe_event_all = __DeviceProxy__unsubscribe_event_all

    DeviceProxy.__str__ = __DeviceProxy__str
    DeviceProxy.__repr__ = __DeviceProxy__str
    DeviceProxy._get_info_ = __DeviceProxy___get_info_

    DeviceProxy.__getattr__ = __safe_call(__DeviceProxy__getattr)
    DeviceProxy.__setattr__ = __safe_call(__DeviceProxy__setattr)
    DeviceProxy.__getitem__ = __safe_call(__DeviceProxy__getitem)
    DeviceProxy.__setitem__ = __safe_call(__DeviceProxy__setitem)
    DeviceProxy.__contains__ = __safe_call(__DeviceProxy__contains)
    DeviceProxy.__dir__ = __safe_call(__DeviceProxy__dir)



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
This is an internal PyTango module.
"""

__all__ = (
    "Util",
    "pyutil_init",
    "TimedAttrData",
    "TimedCmdData",
    "EnsureOmniThread",
    "is_omni_thread",
)

__docformat__ = "restructuredtext"

import os
import sys
import re
import copy

from argparse import ArgumentParser

from tango import AttrQuality
from tango._tango import (
    Util,
    Except,
    DevFailed,
    DbDevInfo,
    EnsureOmniThread,
    is_omni_thread,
    _telemetry,
)

from tango.utils import (
    PyTangoHelpFormatter,
    is_non_str_seq,
    _InterfaceDefinedByIDL,
    _exception_converter,
)
from tango.globals import class_list, cpp_class_list, get_constructed_classes

import collections.abc


class TimedAttrData(_InterfaceDefinedByIDL):
    """This is pure-Python class, which combines both TimedAttrData and AttributeData
    cppTango classes, for use with Util.fill_attr_polling_buffer

    :param value: value to be inserted in polling history. `Default:` :obj:`None`
    :type value: any type compatible with the Tango attribute's dtype

    :param quality: quality of value. `Default:` :obj:`tango.AttrQuality.ATTR_VALID`
    :type quality: :obj:`tango.AttrQuality`

    :param w_value: corresponding written value. Note: should be present only for writable attributes `Default:` :obj:`None`
    :type w_value: any type compatible with the Tango attribute's dtype

    :param error: if the error reading should be inserted. Note: error has a priority over value! `Default:` :obj:`None`
    :type error: :obj:`Exception` or :obj:`tango.DevFailed`

    :param time_stamp: value time stamp in seconds passed since epoch. If not provided, the current system time will be used `Default:` :obj:`None`
    :type time_stamp: :obj:`float`
    """

    def __init__(
        self,
        value=None,
        quality=AttrQuality.ATTR_VALID,
        w_value=None,
        error=None,
        time_stamp=None,
    ):
        self.value = value
        self.quality = quality
        self.w_value = w_value
        self.error = error
        self.time_stamp = time_stamp
        self._initialized = True


class TimedCmdData(_InterfaceDefinedByIDL):
    """This is pure-Python class, which mimics the TimedCmdData
    cppTango class, for use with Util.fill_cmd_polling_buffer

    :param value: value to be inserted in polling history. `Default:` :obj:`None`
    :type value: any type compatible with the Tango commands's dtype

    :param error: if the error reading should be inserted. Note: error has a priority over value! `Default:` :obj:`None`
    :type error: :obj:`Exception` or :obj:`tango.DevFailed`

    :param time_stamp: value time stamp in seconds passed since epoch. If not provided, the current system time will be used `Default:` :obj:`None`
    :type time_stamp: :obj:`float`
    """

    def __init__(
        self,
        value=None,
        error=None,
        time_stamp=None,
    ):
        self.value = value
        self.error = error
        self.time_stamp = time_stamp
        self._initialized = True


def __simplify_device_name(dev_name):
    if dev_name.startswith("tango://"):
        dev_name = dev_name[8:]
    if dev_name.count("/") > 2:
        dev_name = dev_name[dev_name.index("/") + 1 :]
    return dev_name.lower()


#
# Methods on Util
#


def __Util__get_class_list(self):
    """
    get_class_list(self) -> seq<DeviceClass>

            Returns a list of objects of inheriting from DeviceClass

        Parameters : None

        Return     : (seq<DeviceClass>) a list of objects of inheriting from DeviceClass
    """
    return get_constructed_classes()


def __Util__create_device(self, klass_name, device_name, alias=None, cb=None):
    """
    create_device(self, klass_name, device_name, alias=None, cb=None) -> None

        Creates a new device of the given class in the database, creates a new
        DeviceImpl for it and calls init_device (just like it is done for
        existing devices when the DS starts up)

        An optional parameter callback is called AFTER the device is
        registered in the database and BEFORE the init_device for the
        newly created device is called

        Throws tango.DevFailed:
            - the device name exists already or
            - the given class is not registered for this DS.
            - the cb is not a callable

    New in PyTango 7.1.2

    Parameters :
        - klass_name : (str) the device class name
        - device_name : (str) the device name
        - alias : (str) optional alias. Default value is None meaning do not create device alias
        - cb : (callable) a callback that is called AFTER the device is registered
               in the database and BEFORE the init_device for the newly created
               device is called. Typically you may want to put device and/or attribute
               properties in the database here. The callback must receive a parameter
               device_name (str). Default value is None meaning no callback

    Return     : None"""
    if cb is not None and not isinstance(cb, collections.abc.Callable):
        Except.throw_exception(
            "PyAPI_InvalidParameter",
            "The optional cb parameter must be a python callable",
            "Util.create_device",
        )

    db = self.get_database()

    device_name = __simplify_device_name(device_name)

    device_exists = True
    try:
        db.import_device(device_name)
    except DevFailed as df:
        device_exists = not df.args[0].reason == "DB_DeviceNotDefined"

    # 1 - Make sure device name doesn't exist already in the database
    if device_exists:
        Except.throw_exception(
            "PyAPI_DeviceAlreadyDefined",
            f"The device {device_name} is already defined in the database",
            "Util.create_device",
        )

    # 2 - Make sure the device class is known
    klass_list = self.get_class_list()
    klass = None
    for k in klass_list:
        name = k.get_name()
        if name == klass_name:
            klass = k
            break
    if klass is None:
        Except.throw_exception(
            "PyAPI_UnknownDeviceClass",
            f"The device class {klass_name} could not be found",
            "Util.create_device",
        )

    # 3 - Create entry in the database (with alias if necessary)
    dev_info = DbDevInfo()
    dev_info.name = device_name
    dev_info._class = klass_name
    dev_info.server = self.get_ds_name()

    db.add_device(dev_info)

    if alias is not None:
        db.put_device_alias(device_name, alias)

    # from this point on, if anything wrong happens we need to clean the database
    try:
        # 4 - run the callback which tipically is used to initialize
        #     device and/or attribute properties in the database
        if cb is not None:
            cb(device_name)

        # 5 - Initialize device object on this server
        k.device_factory([device_name])
    except Exception:
        try:
            if alias is not None:
                db.delete_device_alias(alias)
        except Exception:
            pass
        db.delete_device(device_name)


def __Util__delete_device(self, klass_name, device_name):
    """
    delete_device(self, klass_name, device_name) -> None

        Deletes an existing device from the database and from this running
        server

        Throws tango.DevFailed:
            - the device name doesn't exist in the database
            - the device name doesn't exist in this DS.

    New in PyTango 7.1.2

    Parameters :
        - klass_name : (str) the device class name
        - device_name : (str) the device name

    Return     : None"""

    db = self.get_database()
    device_name = __simplify_device_name(device_name)
    device_exists = True
    try:
        db.import_device(device_name)
    except DevFailed as df:
        device_exists = not df.args[0].reason == "DB_DeviceNotDefined"

    # 1 - Make sure device name exists in the database
    if not device_exists:
        Except.throw_exception(
            "PyAPI_DeviceNotDefined",
            f"The device {device_name} is not defined in the database",
            "Util.delete_device",
        )

    # 2 - Make sure device name is defined in this server
    class_device_name = f"{klass_name}::{device_name}"
    ds = self.get_dserver_device()
    dev_names = ds.query_device()
    device_exists = False
    for dev_name in dev_names:
        p = dev_name.index("::")
        dev_name = dev_name[:p] + dev_name[p:].lower()
        if dev_name == class_device_name:
            device_exists = True
            break
    if not device_exists:
        Except.throw_exception(
            "PyAPI_DeviceNotDefinedInServer",
            f"The device {class_device_name} is not defined in this server",
            "Util.delete_device",
        )

    db.delete_device(device_name)

    dimpl = self.get_device_by_name(device_name)

    dc = dimpl.get_device_class()
    dc.device_destroyer(device_name)


def __check_arg_for_polling_buffer(history_stack, expected_type, parameter_name):
    # if user gave us just one value - convert it to list
    if not is_non_str_seq(history_stack):
        history_stack = [history_stack]

    for v in history_stack:
        if not isinstance(v, expected_type):
            raise ValueError(
                f"{parameter_name} parameter has type {type(v)}, "
                f"while it must be {expected_type.__name__} object "
                f"or sequence of {expected_type.__name__} objects"
            )
        if isinstance(v.error, Exception):
            v.error = _exception_converter(v.error)
        if isinstance(v.error, DevFailed):
            v.error = v.error.args

    return history_stack


def __Util__fill_attr_polling_buffer(self, device, attribute_name, attr_history_stack):
    """
    fill_attr_polling_buffer(self, device, attribute_name, attr_history_stack) -> None

        Fill attribute polling buffer with your own data. E.g.:

        .. code-block:: python

            def fill_history():
                util = Util.instance(False)
                # note is such case quality will ATTR_VALID, and time_stamp will be time.time()
                util.fill_attr_polling_buffer(device, attribute_name, TimedAttrData(my_new_value))

        or:

        .. code-block:: python

            def fill_history():
                util = Util.instance(False)

                data = TimedAttrData(value=my_new_value,
                                     quality=AttrQuality.ATTR_WARNING,
                                     w_value=my_new_w_value,
                                     time_stamp=my_time)

                util.fill_attr_polling_buffer(device, attribute_name, data)

        or:

        .. code-block:: python

            def fill_history():
                util = Util.instance(False)
                data = [TimedAttrData(my_new_value),
                        TimedAttrData(error=RuntimeError("Cannot read value")]

                util.fill_attr_polling_buffer(device, attribute_name, data)

    :param device: the device to fill attribute polling buffer
    :type device: :obj:`tango.DeviceImpl`

    :param attribute_name: name of the attribute to fill polling buffer
    :type attribute_name: :obj:`str`

    :param attr_history_stack: data to be inserted.
    :type attr_history_stack: :obj:`tango.TimedAttrData` or list[:obj:`tango.TimedAttrData`]

    :return: None

    :raises: :obj:`tango.DevFailed`

    .. versionadded:: 10.1.0
    """

    attr_history_stack = __check_arg_for_polling_buffer(
        attr_history_stack, TimedAttrData, "attr_history_stack"
    )

    self._fill_attr_polling_buffer(device, attribute_name, attr_history_stack)


def __Util__fill_cmd_polling_buffer(self, device, command_name, cmd_history_stack):
    """
    fill_cmd_polling_buffer(self, device, command_name, attr_history_stack) -> None

        Fill attribute polling buffer with your own data. E.g.:


        .. code-block:: python

            def fill_history():
                util = Util.instance(False)
                # note is such time_stamp will be set to time.time()
                util.fill_cmd_polling_buffer(device, command_name, TimedCmdData(my_new_value))

        or:

        .. code-block:: python

            def fill_history():
                util = Util.instance(False)

                data = TimedCmdData(value=my_new_value,
                                     time_stamp=my_time)

                util.fill_cmd_polling_buffer(device, command_name, data)

        or:

        .. code-block:: python

            def fill_history():
                util = Util.instance(False)
                data = [TimedCmdData(my_new_value),
                        TimedCmdData(error=RuntimeError("Cannot read value")]

                util.fill_cmd_polling_buffer(device, command_name, data)


    :param device: the device to fill command polling buffer
    :type device: :obj:`tango.DeviceImpl`

    :param command_name: name of the command to fill polling buffer
    :type command_name: :obj:`str`

    :param cmd_history_stack: data to be inserted
    :type cmd_history_stack: :obj:`tango.TimedCmdData` or list[:obj:`tango.TimedCmdData`]

    :return: None

    :raises: :obj:`tango.DevFailed`

    .. versionadded:: 10.1.0
    """

    cmd_history_stack = __check_arg_for_polling_buffer(
        cmd_history_stack, TimedCmdData, "cmd_history_stack"
    )

    self._fill_cmd_polling_buffer(device, command_name, cmd_history_stack)


def parse_args(args):
    parser = ArgumentParser(
        prog=os.path.splitext(args[0])[0],
        usage="%(prog)s instance_name [-v[trace level]] "
        + "[-host] [-port] [-file=<file_name> | -nodb [-dlist]]",
        add_help=False,
        formatter_class=PyTangoHelpFormatter,
    )

    parser.add_argument("instance_name", nargs="+", help="Device server instance name")
    parser.add_argument(
        "-h", "-?", "--help", action="help", help="show this help message and exit"
    )

    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="count",
        help="set the trace level. "
        + "Can be used in count way: -vv or --verbose --verbose",
    )
    # this option won't be used, since we manually pop all -vN and -v N arguments, but we have to display help about it
    parser.add_argument(
        "-vLEVEL",
        dest="vn",
        action="store",
        metavar=" ",
        help="directly set the trace level to LEVEL",
    )

    parser.add_argument(
        "-file",
        "--file",
        dest="file",
        metavar="FILE_PATH",
        help="start device server using an ASCII file instead of the Tango database",
    )

    parser.add_argument(
        "-host",
        "--host",
        dest="host",
        default="0.0.0.0",
        action="store",
        help="Force the host from which server accepts requests (alternatively use ORBendPoint option)",
    )
    parser.add_argument(
        "-port",
        "--port",
        dest="port",
        default="",
        action="store",
        help="Force the port on which the device server listens (alternatively use ORBendPoint option)",
    )

    if sys.platform.startswith("win"):
        parser.add_argument(
            "-dbg",
            "--dbg",
            dest="dbg",
            action="store_true",
            default=False,
            help="Enable debug",
        )
        parser.add_argument(
            "-i",
            dest="i",
            action="store_true",
            default=False,
            help="Install the service",
        )
        parser.add_argument(
            "-s",
            dest="s",
            action="store_true",
            default=False,
            help="Install the service and choose the automatic startup mode",
        )
        parser.add_argument(
            "-u",
            dest="u",
            action="store_true",
            default=False,
            help="Uninstall the service",
        )

    group = parser.add_argument_group("Run device server without database")
    group.add_argument(
        "-nodb",
        "--nodb",
        dest="nodb",
        action="store_true",
        help="run server without DB",
    )
    group.add_argument(
        "-dlist",
        "--dlist",
        dest="dlist",
        metavar="DEV1,DEV2,etc",
        help="The device name list. This option is supported only with the -nodb option.",
    )

    group = parser.add_argument_group(
        "ORB options (started with -ORBxxx):"
        + "options directly passed to the underlying ORB. Should be rarely used"
    )

    group.add_argument(
        "-ORBendPoint",
        "--ORBendPoint",
        dest="ORBendPoint",
        action="store",
        metavar="giop:tcp:<host>:<port>",
        help="Specifying the host from which server accept "
        "requests and port on which the device server listens.",
    )

    group.add_argument(
        "-ORB<other_option>",
        "--ORB<other_option>",
        dest="ORB_not_used",
        action="store",
        metavar="other_value",
        help="Any other ORB option, e.g., -ORBtraceLevel 5",
    )

    # workaround to add arbitrary ORB options
    for arg in args:
        match = re.match(r"(-ORB|--ORB)(?P<suffix>\w+)", arg)
        if match:
            suffix = match.group("suffix")
            if suffix != "endPoint":
                arg = arg.lstrip("-")
                group.add_argument("-" + arg, "--" + arg, action="store", dest=arg)

    # since -vvvv and -v4 options are incompatible, we have to pop all -vN options
    verbose = None
    for ind, arg in enumerate(args):
        if re.match(r"-[vV][=]?\d+", arg) is not None:
            verbose = int(re.findall(r"\d+", arg)[0])
            args.remove(arg)
            break
        if len(arg) == 2 and re.match(r"-[vV]", arg) is not None:
            if ind + 1 < len(args) and re.match(r"\d+", args[ind + 1]) is not None:
                verbose = int(args[ind + 1])
                args.pop(ind + 1)
                args.remove(arg)
                break

    parsed_args = parser.parse_args(args[1:])

    if parsed_args.port and parsed_args.ORBendPoint is None:
        parsed_args.ORBendPoint = f"giop:tcp:{parsed_args.host:s}:{parsed_args.port:s}"

    if parsed_args.nodb and parsed_args.ORBendPoint is None:
        raise SystemExit(
            "-nodb option should used with [-host] -port or -ORBendPoint options"
        )

    if parsed_args.dlist is not None and not parsed_args.nodb:
        raise SystemExit("-dlist should be used only with -nodb option")

    args = [os.path.splitext(args[0])[0]]

    args += parsed_args.instance_name

    # -v4 has priority on -vvvv
    if verbose is not None:
        args += [f"-v{verbose}"]
    elif parsed_args.verbose is not None:
        args += [f"-v{parsed_args.verbose}"]

    # we add back only exist options
    for key, value in parsed_args.__dict__.items():
        if type(value) is bool:
            if value:
                args += [f"-{key:s}"]
        elif value is not None:
            if key == "file":
                args += [f"-{key:s}={value:s}"]
            elif key not in [
                "host",
                "port",
                "verbose",
                "instance_name",
                "ORB_not_used",
            ]:
                args += [f"-{key:s}", f"{value:s}"]

    return args


def __Util__init__(self, args):
    args = parse_args(copy.copy(args))
    Util.__init_orig__(self, args)


def __Util__init(args):
    args = parse_args(list(args))
    return Util.__init_orig(args)


def __Util__add_TgClass(self, klass_device_class, klass_device, device_class_name=None):
    """Register a new python tango class. Example::

        util.add_TgClass(MotorClass, Motor)
        util.add_TgClass(MotorClass, Motor, 'Motor') # equivalent to previous line

    .. deprecated:: 7.1.2
        Use :meth:`tango.Util.add_class` instead."""
    if device_class_name is None:
        device_class_name = klass_device.__name__
    class_list.append((klass_device_class, klass_device, device_class_name))


def __Util__add_Cpp_TgClass(self, device_class_name, tango_device_class_name):
    """Register a new C++ tango class.

    If there is a shared library file called MotorClass.so which
    contains a MotorClass class and a _create_MotorClass_class method. Example::

        util.add_Cpp_TgClass('MotorClass', 'Motor')

    .. note:: the parameter 'device_class_name' must match the shared
              library name.

    .. deprecated:: 7.1.2
        Use :meth:`tango.Util.add_class` instead."""
    cpp_class_list.append((device_class_name, tango_device_class_name))


def __Util__add_class(self, *args, **kwargs):
    """
    add_class(self, class<DeviceClass>, class<DeviceImpl>, language="python") -> None

        Register a new tango class ('python' or 'c++').

        If language is 'python' then args must be the same as
        :meth:`tango.Util.add_TgClass`. Otherwise, args should be the ones
        in :meth:`tango.Util.add_Cpp_TgClass`. Example::

            util.add_class(MotorClass, Motor)
            util.add_class('CounterClass', 'Counter', language='c++')

    New in PyTango 7.1.2"""
    language = kwargs.get("language", "python")
    f = self.add_TgClass
    if language != "python":
        f = self.add_Cpp_TgClass
    return f(*args)


def __init_Util():
    Util.__init_orig__ = Util.__init__
    Util.__init__ = __Util__init__
    Util.__init_orig = staticmethod(Util.init)
    Util.init = staticmethod(__Util__init)
    Util.add_TgClass = __Util__add_TgClass
    Util.add_Cpp_TgClass = __Util__add_Cpp_TgClass
    Util.add_class = __Util__add_class
    Util.get_class_list = __Util__get_class_list
    Util.create_device = __Util__create_device
    Util.delete_device = __Util__delete_device
    Util.fill_attr_polling_buffer = __Util__fill_attr_polling_buffer
    Util.fill_cmd_polling_buffer = __Util__fill_cmd_polling_buffer


#
# EnsureOmniThread context handler
#


def __EnsureOmniThread__enter__(self):
    self._acquire()
    return self


def __EnsureOmniThread__exit__(self, exc_type, exc_value, traceback):
    self._release()
    return False


def __init_EnsureOmniThread():
    EnsureOmniThread.__enter__ = __EnsureOmniThread__enter__
    EnsureOmniThread.__exit__ = __EnsureOmniThread__exit__


#
# TraceContextScope context handler
#


def __TraceContextScope__enter__(self):
    self._acquire()
    return self


def __TraceContextScope__exit__(self, exc_type, exc_value, traceback):
    self._release()
    return False


def __init_TraceContextScope():
    _telemetry.TraceContextScope.__enter__ = __TraceContextScope__enter__
    _telemetry.TraceContextScope.__exit__ = __TraceContextScope__exit__


def pyutil_init(doc=True):
    __init_Util()
    __init_EnsureOmniThread()
    __init_TraceContextScope()



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
This is an internal PyTango module.
"""

__all__ = (
    "get_class",
    "get_classes",
    "get_cpp_class",
    "get_cpp_classes",
    "get_constructed_class",
    "get_constructed_classes",
    "class_factory",
    "delete_class_list",
    "class_list",
    "cpp_class_list",
    "constructed_class",
)

__docformat__ = "restructuredtext"

# list of tuple<DeviceClass class, DeviceImpl class, tango device class name>
class_list = []

# list of tuple<DeviceClass name, tango device class name>
cpp_class_list = []

# list of DeviceClass objects, one for each registered device class
constructed_class = []


def get_classes():
    global class_list
    return class_list


def get_class(name):
    for klass_info in get_classes():
        if klass_info[2] == name:
            return klass_info
    return None


def get_class_by_class(klass):
    for klass_info in get_classes():
        if klass_info[0] == klass:
            return klass_info
    return None


def get_cpp_classes():
    global cpp_class_list
    return cpp_class_list


def get_cpp_class(name):
    for klass_info in get_cpp_classes():
        if klass_info[1] == name:
            return klass_info
    return None


def get_constructed_classes():
    global constructed_class
    return constructed_class


def get_constructed_class(name):
    for klass in get_constructed_classes():
        if klass.get_name() == name:
            return klass
    return None


def get_constructed_class_by_class(klass):
    for k in get_constructed_classes():
        if k.__class__ == klass:
            return k
    return None


#
# A method to delete Tango classes from Python
#


def delete_class_list():
    global constructed_class
    if len(constructed_class) != 0:
        del constructed_class[:]


#
# A generic class_factory method
#


def class_factory():
    local_class_list = get_classes()
    local_cpp_class_list = get_cpp_classes()

    if (len(local_class_list) + len(local_cpp_class_list)) == 0:
        print("Oups, no Tango class defined within this device server !!!")
        print("Sorry, but I exit")
        import sys

        sys.exit()

    # Call the delete_class_list function in order to clear the global
    # constructed class Python list. This is necessary only in case of
    # ServerRestart command
    delete_class_list()

    # Clear also all remaining references to the old DeviceClass instances
    for class_info in local_class_list:
        class_info[1]._device_class_instance = None

    local_constructed_class = get_constructed_classes()
    for class_info in local_class_list:
        device_class_class = class_info[0]
        tango_device_class_name = class_info[2]
        device_class = device_class_class(tango_device_class_name)
        local_constructed_class.append(device_class)



#!/usr/bin/env python
# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
Clock Device server showing how to write a TANGO server with a Clock device
which has attributes:

  - time: read-only scalar float
  - gmtime: read-only sequence (spectrum) of integers
  - noon:  read-only enumerated type

commands:

  - ctime: in: float parameter; returns a string
  - mktime: in: sequence (spectrum) of 9 integers; returns a float

Requires at least PyTango 9.5.0 for type hinting to work.  On earlier
versions, use something like `@attribute(dtype=Noon)`.
"""

import time
from enum import IntEnum

from tango.server import Device, attribute, command


class Noon(IntEnum):
    AM = 0  # DevEnum's must start at 0
    PM = 1  # and increment by 1


class DisplayType(IntEnum):
    ANALOG = 0  # DevEnum's must start at 0
    DIGITAL = 1  # and increment by 1


class Clock(Device):
    display_type = DisplayType.ANALOG

    @attribute
    def time(self) -> float:
        return time.time()

    @attribute(max_dim_x=9)
    def gmtime(self) -> tuple[int]:
        return time.gmtime()

    @attribute
    def noon(self) -> Noon:
        time_struct = time.gmtime(time.time())
        return Noon.AM if time_struct.tm_hour < 12 else Noon.PM

    @attribute
    def display(self) -> DisplayType:
        return self.display_type

    @display.setter
    def display(self, display_type: int):
        # note that we receive an integer, not an enum instance,
        # so we have to convert that to an instance of our enum.
        self.display_type = DisplayType(display_type)

    @command(dtype_in=float, dtype_out=str)
    def ctime(self, seconds):
        """
        Convert a time in seconds since the Epoch to a string in local time.
        This is equivalent to asctime(localtime(seconds)). When the time tuple
        is not present, current time as returned by localtime() is used.
        """
        return time.ctime(seconds)

    @command
    def mktime(self, tupl: tuple[int]) -> float:
        return time.mktime(tuple(tupl))


if __name__ == "__main__":
    Clock.run_server()



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
import json
from PyTango import DevState
from PyTango.server import Device
from PyTango.server import attribute, command, device_property


class TuringMachine(Device):
    blank_symbol = device_property(dtype=str, default_value=" ")
    initial_state = device_property(dtype=str, default_value="init")

    def init_device(self):
        Device.init_device(self)
        self.__tape = {}
        self.__head = 0
        self.__state = self.initial_state
        self.__final_states = []
        self.__transition_function = None
        self.set_state(DevState.RUNNING)

    @attribute(dtype=(str,))
    def final_states(self):
        return self.__final_states

    @final_states.write
    def final_states(self, final_states):
        self.__final_states = final_states

    @attribute(dtype=str)
    def transition_function(self):
        return self.__transition_function

    @transition_function.write
    def transition_function(self, func_str):
        self.__transition_function = tf = {}
        for k, v in json.loads(func_str).items():
            tf[tuple(str(k).split(","))] = map(str, v)
        print(tf)

    @attribute(dtype=str)
    def tape(self):
        s, keys = "", self.__tape.keys()
        min_used, max_used = min(keys), max(keys)
        for i in range(min_used, max_used):
            s += self.__tape.get(i, self.__blank_symbol)
        return s

    @command
    def step(self):
        char_under_head = self.__tape.get(self.__head, self.blank_symbol)
        x = self.__state, char_under_head
        if x in self.__transition_function:
            y = self.__transition_function[x]
            self.__tape[self.__head] = y[1]
            if y[2] == "R":
                self.__head += 1
            elif y[2] == "L":
                self.__head -= 1
            self.__state = y[0]
        print(self.__state)

    def dev_state(self):
        if self.__state in self.__final_states:
            return DevState.ON
        else:
            return DevState.RUNNING


if __name__ == "__main__":
    TuringMachine.run_server()



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""Server for tango objects."""

import os
import sys
import logging
import weakref
from inspect import getfullargspec
import inspect
import functools
import numpy

try:
    from warnings import deprecated
except ImportError:
    from typing_extensions import deprecated

from tango.codec import loads, dumps
from tango.attr_data import AttrData
from tango.utils import TO_TANGO_TYPE
from tango import AttrDataFormat, CmdArgType, GreenMode
from tango import DbDevInfo, Database, DevState
from tango.server import Device, _to_classes, _add_classes
from tango.device_server import get_worker, set_worker
from tango.green import get_executor

__all__ = ("Server",)

_CLEAN_UP_TEMPLATE = """
import sys
from tango import Database

db = Database()
server_instance = '{server_instance}'
try:
    devices = db.get_device_class_list(server_instance)[::2]
    for device in devices:
        db.delete_device(device)
        try:
            db.delete_device_alias(db.get_alias(device))
        except Exception:
            pass
except Exception:
    print ('Failed to cleanup!')
"""


def __to_tango_type_fmt(value):
    dfmt = AttrDataFormat.SCALAR
    value_t = type(value)
    dtype = TO_TANGO_TYPE.get(value_t)
    max_dim_x, max_dim_y = 1, 0
    if dtype is None:
        if isinstance(value, numpy.ndarray):
            dtype = TO_TANGO_TYPE.get(value.dtype.name)
            shape_l = len(value.shape)
            if shape_l == 1:
                dfmt = AttrDataFormat.SPECTRUM
                max_dim_x = max(2**16, value.shape[0])
            elif shape_l == 2:
                dfmt = AttrDataFormat.IMAGE
                max_dim_x = max(2**16, value.shape[0])
                max_dim_y = max(2**16, value.shape[1])
        else:
            dtype = CmdArgType.DevEncoded
    return dtype, dfmt, max_dim_x, max_dim_y


@deprecated(
    "create_tango_class function was an experimental API - scheduled for removal in PyTango 11.0.0"
)
def create_tango_class(server, obj, tango_class_name=None, member_filter=None):
    slog = server.server_instance.replace("/", ".")
    log = logging.getLogger("tango.Server." + slog)

    obj_klass = obj.__class__
    obj_klass_name = obj_klass.__name__

    if tango_class_name is None:
        tango_class_name = obj_klass_name

    class DeviceDispatcher(Device):
        TangoClassName = tango_class_name

        def __init__(self, tango_class_obj, name):
            tango_object = server.get_tango_object(name)
            self.__tango_object = weakref.ref(tango_object)
            Device.__init__(self, tango_class_obj, name)

        def init_device(self):
            Device.init_device(self)
            self.set_state(DevState.ON)

        @property
        def _tango_object(self):
            return self.__tango_object()

        @property
        def _object(self):
            return self._tango_object._object

    DeviceDispatcher.__name__ = tango_class_name
    DeviceDispatcher.TangoClassName = tango_class_name
    DeviceDispatcherClass = DeviceDispatcher.TangoClassClass

    for name in dir(obj):
        if name.startswith("_"):
            continue
        log.debug("inspecting %s.%s", obj_klass_name, name)
        try:
            member = getattr(obj, name)
        except Exception:
            log.info("failed to inspect member '%s.%s'", obj_klass_name, name)
            log.debug("Details:", exc_info=1)
        if inspect.isclass(member) or inspect.ismodule(member):
            continue
        if member_filter and not member_filter(obj, tango_class_name, name, member):
            log.debug("filtered out %s from %s", name, tango_class_name)
            continue
        if inspect.isroutine(member):
            # try to find out if there are any parameters
            in_type = CmdArgType.DevEncoded
            out_type = CmdArgType.DevEncoded
            try:
                arg_spec = getfullargspec(member)
                if not arg_spec.args:
                    in_type = CmdArgType.DevVoid
            except TypeError:
                pass

            if in_type == CmdArgType.DevVoid:

                def _command(dev, func_name=None):
                    obj = dev._object
                    f = getattr(obj, func_name)
                    result = server.worker.execute(f)
                    return server.dumps(result)

            else:

                def _command(dev, param, func_name=None):
                    obj = dev._object
                    args, kwargs = loads(*param)
                    f = getattr(obj, func_name)
                    result = server.worker.execute(f, *args, **kwargs)
                    return server.dumps(result)

            cmd = functools.partial(_command, func_name=name)
            cmd.__name__ = name
            doc = member.__doc__
            if doc is None:
                doc = ""
            cmd.__doc__ = doc
            setattr(DeviceDispatcher, name, cmd)
            DeviceDispatcherClass.cmd_list[name] = [[in_type, doc], [out_type, ""]]
        else:
            read_only = False
            if hasattr(obj_klass, name):
                kmember = getattr(obj_klass, name)
                if inspect.isdatadescriptor(kmember):
                    if kmember.fset is None:
                        read_only = True
                else:
                    continue
            value = member
            dtype, fmt, x, y = __to_tango_type_fmt(value)
            if dtype is None or dtype == CmdArgType.DevEncoded:
                dtype = CmdArgType.DevEncoded
                fmt = AttrDataFormat.SCALAR

                def read(dev, attr):
                    name = attr.get_name()
                    value = server.worker.execute(getattr, dev._object, name)
                    attr.set_value(*server.dumps(value))

                def write(dev, attr):
                    name = attr.get_name()
                    value = attr.get_write_value()
                    value = loads(*value)
                    server.worker.execute(setattr, dev._object, name, value)

            else:

                def read(dev, attr):
                    name = attr.get_name()
                    value = server.worker.execute(getattr, dev._object, name)
                    attr.set_value(value)

                def write(dev, attr):
                    name = attr.get_name()
                    value = attr.get_write_value()
                    server.worker.execute(setattr, dev._object, name, value)

            read.__name__ = "_read_" + name
            setattr(DeviceDispatcher, read.__name__, read)

            pars = dict(
                name=name, dtype=dtype, dformat=fmt, max_dim_x=x, max_dim_y=y, fget=read
            )
            if not read_only:
                write.__name__ = "_write_" + name
                pars["fset"] = write
                setattr(DeviceDispatcher, write.__name__, write)
            attr_data = AttrData.from_dict(pars)
            DeviceDispatcherClass.attr_list[name] = attr_data
    return DeviceDispatcher


@deprecated(
    "Server class was an experimental API - scheduled for removal in PyTango 11.0.0"
)
class Server:
    """
    Server helper
    """

    Phase0, Phase1, Phase2 = range(3)
    PreInitPhase = Phase1
    PostInitPhase = Phase2

    class TangoObjectAdapter:
        def __init__(self, server, obj, full_name, alias=None, tango_class_name=None):
            self.__server = weakref.ref(server)
            self.full_name = full_name
            self.alias = alias
            self.class_name = obj.__class__.__name__
            if tango_class_name is None:
                tango_class_name = self.class_name
            self.tango_class_name = tango_class_name
            self.__object = weakref.ref(obj, self.__onObjectDeleted)

        def __onObjectDeleted(self, object_weak):
            self.__object = None
            server = self._server
            server.log.info("object deleted %s(%s)", self.class_name, self.full_name)
            server.unregister_object(self.full_name)

        @property
        def _server(self):
            return self.__server()

        @property
        def _object(self):
            obj = self.__object
            if obj is None:
                return None
            return obj()

    def __init__(
        self,
        server_name,
        server_type=None,
        port=None,
        event_loop_callback=None,
        init_callbacks=None,
        auto_clean=False,
        green_mode=None,
        tango_classes=None,
        protocol="pickle",
    ):
        if server_name is None:
            raise ValueError("Must give a valid server name")
        self.__server_name = server_name
        self.__server_type = server_type
        self.__port = port
        self.__event_loop_callback = event_loop_callback
        if init_callbacks is None:
            init_callbacks = {}
        self.__init_callbacks = init_callbacks
        self.__util = None
        self.__objects = {}
        self.__running = False
        self.__auto_clean = auto_clean
        self.__green_mode = green_mode
        self.__protocol = protocol
        self.__tango_classes = _to_classes(tango_classes or [])
        self.__tango_devices = []
        if self.async_mode:
            self.__worker = get_executor(self.green_mode)
        else:
            self.__worker = get_worker()
        set_worker(self.__worker)
        self.log = logging.getLogger("tango.Server")
        self.__phase = Server.Phase0

    def __build_args(self):
        args = [self.server_type, self.__server_name]
        if self.__port is not None:
            args.extend(["-ORBendPoint", f"giop:tcp:0.0.0.0:{self.__port}"])
        return args

    def __exec_cb(self, cb):
        if not cb:
            return
        self.worker.execute(cb)

    def __find_tango_class(self, key):
        pass

    def __prepare(self):
        """Update database with existing devices"""
        self.log.debug("prepare")

        if self.__phase > 0:
            raise RuntimeError("Internal error: Can only prepare in phase 0")

        server_instance = self.server_instance
        db = Database()

        # get list of server devices if server was already registered
        server_registered = server_instance in db.get_server_list()

        if server_registered:
            dserver_name = f"dserver/{server_instance}"
            if db.import_device(dserver_name).exported:
                import tango

                dserver = tango.DeviceProxy(dserver_name)
                try:
                    dserver.ping()
                    raise Exception("Server already running")
                except Exception:
                    self.log.info("Last time server was not properly shutdown!")
            _, db_device_map = self.get_devices()
        else:
            db_device_map = {}

        db_devices_add = {}

        # all devices that are registered in database that are not registered
        # as tango objects or for which the tango class changed will be removed
        db_devices_remove = set(db_device_map) - set(self.__objects)

        for local_name, local_object in self.__objects.items():
            local_class_name = local_object.tango_class_name
            db_class_name = db_device_map.get(local_name)
            if db_class_name:
                if local_class_name != db_class_name:
                    db_devices_remove.add(local_name)
                    db_devices_add[local_name] = local_object
            else:
                db_devices_add[local_name] = local_object

        for device in db_devices_remove:
            db.delete_device(device)
            try:
                db.delete_device_alias(db.get_alias(device))
            except Exception:
                pass

        # register devices in database

        # add DServer
        db_dev_info = DbDevInfo()
        db_dev_info.server = server_instance
        db_dev_info._class = "DServer"
        db_dev_info.name = "dserver/" + server_instance

        db_dev_infos = [db_dev_info]
        aliases = []
        for obj_name, obj in db_devices_add.items():
            db_dev_info = DbDevInfo()
            db_dev_info.server = server_instance
            db_dev_info._class = obj.tango_class_name
            db_dev_info.name = obj.full_name
            db_dev_infos.append(db_dev_info)
            if obj.alias:
                aliases.append((obj.full_name, obj.alias))

        db.add_server(server_instance, db_dev_infos)

        # add aliases
        for alias_info in aliases:
            db.put_device_alias(*alias_info)

    def __clean_up_process(self):
        if not self.__auto_clean:
            return
        clean_up = _CLEAN_UP_TEMPLATE.format(server_instance=self.server_instance)
        import subprocess

        res = subprocess.call([sys.executable, "-c", clean_up])
        if res:
            self.log.error("Failed to cleanup")

    def __initialize(self):
        self.log.debug("initialize")
        async_mode = self.async_mode
        event_loop = self.__event_loop_callback

        util = self.tango_util
        u_instance = util.instance()

        if async_mode:
            if event_loop:
                event_loop = functools.partial(self.worker.execute, event_loop)
        if event_loop:
            u_instance.server_set_event_loop(event_loop)

        _add_classes(util, self.__tango_classes)

    def __run(self, timeout=None):
        return self.worker.run(self.__tango_loop, wait=True, timeout=timeout)

    def __tango_loop(self):
        self.log.debug("server loop started")
        self.__running = True
        u_instance = self.tango_util.instance()
        u_instance.server_init()
        self._phase = Server.Phase2
        self.log.info("Ready to accept request")
        u_instance.server_run()
        if self.__auto_clean:
            self.__clean_up_process()
        self.log.debug("server loop exit")

    @property
    def _phase(self):
        return self.__phase

    @_phase.setter
    def _phase(self, phase):
        self.__phase = phase
        cb = self.__init_callbacks.get(phase)
        self.__exec_cb(cb)

    @property
    def server_type(self):
        server_type = self.__server_type
        if server_type is None:
            server_file = os.path.basename(sys.argv[0])
            server_type = os.path.splitext(server_file)[0]
        return server_type

    @property
    def server_instance(self):
        return f"{self.server_type}/{self.__server_name}"

    @property
    def tango_util(self):
        if self.__util is None:
            import tango

            self.__util = tango.Util(self.__build_args())
            self._phase = Server.Phase1
        return self.__util

    @property
    def green_mode(self):
        gm = self.__green_mode
        if gm is None:
            from tango import get_green_mode

            gm = get_green_mode()
        return gm

    @green_mode.setter
    def green_mode(self, gm):
        if gm == self.__green_mode:
            return
        if self.__running:
            raise RuntimeError("Cannot change green mode while server is running")
        self.__green_mode = gm

    @property
    def async_mode(self):
        return self.green_mode in (GreenMode.Gevent, GreenMode.Asyncio)

    @property
    def worker(self):
        return self.__worker

    def dumps(self, obj):
        return dumps(self.__protocol, obj)

    def get_devices(self):
        """
        Helper that retuns a dict of devices for this server.

        :return:
            Returns a tuple of two elements:
              - dict<tango class name : list of device names>
              - dict<device names : tango class name>
        :rtype: tuple<dict, dict>
        """
        if self.__util is None:
            import tango

            db = tango.Database()
        else:
            db = self.__util.get_database()
        server = self.server_instance
        dev_list = db.get_device_class_list(server)
        class_map, dev_map = {}, {}
        for class_name, dev_name in zip(dev_list[1::2], dev_list[::2]):
            dev_names = class_map.get(class_name)
            if dev_names is None:
                class_map[class_name] = dev_names = []
            dev_name = dev_name.lower()
            dev_names.append(dev_name)
            dev_map[dev_name] = class_name
        return class_map, dev_map

    def get_tango_object(self, name):
        return self.__objects.get(name.lower())

    def get_tango_class(self, tango_class_name):
        for klass in self.__tango_classes:
            if klass.TangoClassName == tango_class_name:
                return klass

    def register_tango_device(self, klass, name):
        if inspect.isclass(klass):
            if isinstance(klass, Device):
                # TODO
                raise NotImplementedError
            else:
                raise ValueError
        else:
            raise NotImplementedError

    def register_tango_class(self, klass):
        if self._phase > Server.Phase1:
            raise RuntimeError(
                "Cannot add new class after phase 1 (i.e. after server_init)"
            )
        self.__tango_classes.append(klass)

    def unregister_object(self, name):
        tango_object = self.__objects.pop(name.lower())
        if self._phase > Server.Phase1:
            import tango

            util = tango.Util.instance()
            if not util.is_svr_shutting_down():
                util.delete_device(tango_object.tango_class_name, name)

    def register_object(self, obj, name, tango_class_name=None, member_filter=None):
        """
        :param member_filter:
            callable(obj, tango_class_name, member_name, member) -> bool
        """
        slash_count = name.count("/")
        if slash_count == 0:
            alias = name
            full_name = f"{self.server_instance}/{name}"
        elif slash_count == 2:
            alias = None
            full_name = name
        else:
            raise ValueError("Invalid name")

        class_name = tango_class_name or obj.__class__.__name__
        tango_class = self.get_tango_class(class_name)

        if tango_class is None:
            tango_class = create_tango_class(
                self, obj, class_name, member_filter=member_filter
            )
            self.register_tango_class(tango_class)

        tango_object = self.TangoObjectAdapter(
            self, obj, full_name, alias, tango_class_name=class_name
        )
        self.__objects[full_name.lower()] = tango_object
        if self._phase > Server.Phase1:
            import tango

            util = tango.Util.instance()
            util.create_device(class_name, name)
        return tango_object

    def run(self, timeout=None):
        self.log.debug("run")
        async_mode = self.async_mode
        running = self.__running
        if not running:
            self.__prepare()
            self.__initialize()
        else:
            if not async_mode:
                raise RuntimeError("Server is already running")
        self.__run(timeout=timeout)



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
"""
Basic unit test for a PowerSupply device.  Requires pytest.
"""

from tango.test_utils import DeviceTestContext

from ps0a import PowerSupply


def test_calibrate():
    """Test device calibration and voltage reading."""
    with DeviceTestContext(PowerSupply, process=True) as proxy:
        proxy.calibrate()
        assert proxy.voltage == 1.5



#!/usr/bin/env python3
# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""Trivial example of a low-level API device server."""

import sys
import tango


class Motor(tango.Device_6Impl):
    def __init__(self, cl, name):
        tango.Device_6Impl.__init__(self, cl, name)
        Motor.init_device(self)

    def delete_device(self):
        pass

    def init_device(self):
        self.get_device_properties(self.get_device_class())
        self.attr_position_read = 1.0

    def always_executed_hook(self):
        pass

    def read_position(self, attr):
        attr.set_value(self.attr_position_read)

    def read_attr_hardware(self, data):
        pass


class MotorClass(tango.DeviceClass):
    class_property_list = {}
    device_property_list = {}
    cmd_list = {}
    attr_list = {"position": [[tango.DevDouble, tango.SCALAR, tango.READ]]}

    def __init__(self, name):
        tango.DeviceClass.__init__(self, name)
        self.set_type(name)


def main():
    try:
        util = tango.Util(sys.argv)
        util.add_class(MotorClass, Motor, "Motor")
        U = tango.Util.instance()
        U.server_init()
        U.server_run()
    except tango.DevFailed as exc:
        print(f"-------> Received a DevFailed exception: {exc}")
    except Exception as exc:
        print(f"-------> An unforeseen exception occurred: {exc}")


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
import tango
from tango.server import Device
from tango.server import attribute

__all__ = "FwdServer"


class FwdServer(Device):
    """
    Start this server: python FwdServer.py myFwdServer
    Start the server containing the root attribute.
    Then using jive select the Attribute properties from myFwdServer
    Select you forwarded attribute and add the value to __root_att
    e.g. __root_att   ->  x/y/z/root_attribute_name
    Now restart the FwdServer
    """

    def init_device(self):
        Device.init_device(self)
        self._current = 0.0
        self.set_state(tango.DevState.ON)

    voltage = attribute(forwarded=True)

    @attribute(label="Current", dtype="float")
    def current(self):
        return self._current


# ----------
# Run server
# ----------


if __name__ == "__main__":
    FwdServer.run_server()
